/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package math;

import channel.Channel;
import error.ErrorLogger;
import java.io.FileInputStream;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Locale;
import java.util.Properties;
import error.Error;

/**
 * Class for performing various math functions.
 * @author Lauritz Møllersen
 */
public abstract class MathFunction {

    /**
     * Method to get the average of two channels
     * @param chan1 Channel 1
     * @param chan2 Channel 2
     * @param name Name of the new channel
     * @return The new channel
     */
    public static Channel averageChan(Channel chan1, Channel chan2, String name) {
        double[] values1 = chan1.getCalibratedValues();
        double[] values2 = chan2.getCalibratedValues();
        final int size = values1.length;
        Channel tempChannel = null;

        if (size == values2.length) {
            double[] tempValues = new double[size];
            double v1;
            double v2;
            for (int i = 0; i < size; i++) {
                v1 = values1[i];
                v2 = values2[i];
                tempValues[i] = (v1 + v2) / 2.0;
            }
            tempChannel = new Channel(tempValues, name);
            tempChannel.setCalibratedValues(tempValues);
            tempChannel.setType(parser.Constants.CHAN_CREATED);
        }
        return tempChannel;
    }

    /**
     * Method to derivate a channel.
     * @param channel The {@link Channel} to be derived.
     * @param name Name of the new channel
     * @return Finished derived channel.
     */
    public static Channel derivate(Channel channel, String name) {
        double[] values = channel.getCalibratedValues();
        final int size = values.length;
        double[] tempValues = new double[size];
        tempValues[0] = 0.0;
        for (int i = 0; i < size - 1; i++) {
            tempValues[i + 1] = values[i + 1] - values[i];
        }
        return new Channel(tempValues, name);
    }

    /**
     * Method to add a value to all values in a channel.
     * @param chan1 Channel to add to
     * @param chan2 Channel to add
     * @param name Name of the new channel
     * @return The new channel
     */
    public static Channel add(Channel chan1, Channel chan2, String name) {
        double[] values1 = chan1.getCalibratedValues();
        double[] values2 = chan2.getCalibratedValues();
        final int size;
        if (values1.length == values2.length) {
            size = values1.length;
        } else {
            ErrorLogger.addError(new Error("Math channel error", "Error adding " + chan1.getName() + " and " + 
                    chan2.getName() + ". Check values and sizes."));
            return null;
        }
        double[] tempValues = new double[size];
        for (int i = 0; i < size; i++) {
            tempValues[i] = values1[i] + values2[i];
        }
        return new Channel(tempValues, name);
    }

    /**
     * Method to subtract a value to all values in a channel.
     * @param chan1 Channel to subtract from
     * @param chan2 Channel to subtract
     * @param name Name of the new channel
     * @return The new channel
     */
    public static Channel subtract(Channel chan1, Channel chan2, String name) {
        double[] values1 = chan1.getCalibratedValues();
        double[] values2 = chan2.getCalibratedValues();
        final int size;
        if (values1.length == values2.length) {
            size = values1.length;
        } else {
            ErrorLogger.addError(new Error("Math channel error", "Error subtracting " + chan1.getName() + " and " + 
                    chan2.getName() + ". Check values and sizes."));
            return null;
        }
        double[] tempValues = new double[size];
        for (int i = 0; i < size; i++) {
            tempValues[i] = values1[i] - values2[i];
        }
        return new Channel(tempValues, name);
    }

    /**
     * Method to divide all values in a channel with a value.
     * @param chan1 Channel to divide
     * @param chan2 Channel to divide with
     * @param name Name of the new channel
     * @return The new channel
     */
    public static Channel divide(Channel chan1, Channel chan2, String name) {
        double[] values1 = chan1.getCalibratedValues();
        double[] values2 = chan2.getCalibratedValues();
        final int size;
        if (values1.length == values2.length) {
            size = values1.length;
        } else {
            ErrorLogger.addError(new Error("Math channel error", "Error dividing " + chan1.getName() + " and " + 
                    chan2.getName() + ". Check values and sizes."));
            return null;
        }
        double[] tempValues = new double[size];
        for (int i = 0; i < size; i++) {
            tempValues[i] = values1[i] / values2[i];
        }
        return new Channel(tempValues, name);
    }

    /**
     * Method to multiply all values in a channel with a value.
     * @param chan1 Channel to multiply
     * @param chan2 Channel to multiply with
     * @param name Name of the new channel
     * @return The new channel
     */
    public static Channel multiply(Channel chan1, Channel chan2, String name) {
        double[] values1 = chan1.getCalibratedValues();
        double[] values2 = chan2.getCalibratedValues();
        final int size;
        if (values1.length == values2.length) {
            size = values1.length;
        } else {
            ErrorLogger.addError(new Error("Math channel error", "Error multiplying " + chan1.getName() + " and " + 
                    chan2.getName() + ". Check values and sizes."));
            return null;
        }
        double[] tempValues = new double[size];
        for (int i = 0; i < size; i++) {
            tempValues[i] = values1[i] * values2[i];
        }
        return new Channel(tempValues, name);
    }

    /**
     * Method to generate a race track map using the lateral G-force values and the speed.
     * @param latG The lateral G-force values.
     * @param speed The speed values.
     * @param hz The hz (sampling rate) of the values.
     * @param unitFactor The conversion factor from the speed values to m/s.
     * @param startOffset 
     * @param endOffset 
     * @return A map with x and y values in arrays, for drawing.
     */
    public static Map calculateMap(double[] latG, double[] speed, int hz, double unitFactor, int startOffset, int endOffset) {
        final int size = endOffset - startOffset;
        // need to find what offset to start at and where to end
        double[] positionX = new double[size];
        double[] positionY = new double[size];
        double tempSpeed;
        double r;
        double dA;
        double A = 0.0;
        double Vx;
        double Vy;
        double X0 = 0.0;
        double Y0 = 0.0;

        for (int i = 0; i < size; i++) {
            tempSpeed = speed[i + startOffset] * unitFactor;
            r = (tempSpeed * tempSpeed) / (latG[i + startOffset] * 9.81);
            dA = (tempSpeed / r) / hz;
            Vx = tempSpeed * Math.cos(A);
            Vy = tempSpeed * Math.sin(A);
            A = A + dA;

            X0 = X0 + (Vx / hz);
            Y0 = Y0 + (Vy / hz);
            positionX[i] = X0;
            positionY[i] = Y0;
        }

        Map map = new Map(positionX, positionY);
        map = smoothMap(map);
        return map;
    }

    // to smooth out small errors such as the car drifting (which corrupts the g-force measurements)
    private static Map smoothMap(Map map) {
        double[] x = map.getPositionX();
        double[] y = map.getPositionY();
        final int length = x.length;
        final double diffX = x[length - 1] - x[0];
        final double diffY = y[length - 1] - y[0];
        final double addX = diffX / length;
        final double addY = diffY / length;
        for (int i = 0; i < length; i++) {
            x[i] += diffX - i * addX;
            y[i] += diffY - i * addY;
        }
        return new Map(x, y);
    }

    /**
     * Method to find the hz based on the time-Channel
     * @param time Channel containing the time-values
     * @return The hz (amount of samples per second) in the data
     */
    public static int findHz(Channel time) {
        double[] values = time.getValues();
        double temp = 1 / values[1] - values[0];
        int tempInt = (int) temp;

        // rounding up to a tenth
        tempInt = tempInt + 5;
        tempInt = tempInt / 10;
        tempInt = tempInt * 10;

        return tempInt;
    }

    /**
     * Small method to format decimal numbers from many decimals to maximum two.
     * @param d The decimal number.
     * @return Decimal number with only two decimals.
     */
    public static double roundTwoDecimals(double d) {
        DecimalFormatSymbols otherSymbols = new DecimalFormatSymbols(Locale.GERMAN);
        otherSymbols.setDecimalSeparator('.');
        otherSymbols.setGroupingSeparator(',');
        DecimalFormat df = new DecimalFormat("#.##", otherSymbols);
        try {
            return Double.valueOf(df.format(d));
        } catch (Exception e) {
            error.ErrorLogger.addError(new error.Error(e, 0, "Error rounding decimal", "Wrong input"));
            return Double.NaN;
        }
    }

    /**
     * Method to read Java Properties from a given file
     * @param path The path of the file
     * @return The Properties
     */
    public static Properties readProperties(String path) {
        FileInputStream fs;
        Properties prop = new Properties();
        try {
            fs = new FileInputStream(path);
            prop.load(fs);
            fs.close();
        } catch (Exception e) {
            ErrorLogger.addError(new Error(e, 1, "Error loading calibration properties", "There is an error within the calibration file"));
        }
        return prop;
    }
}