/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package parser;

import channel.*;
import error.ErrorLogger;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import error.Error;
import java.util.HashMap;
import java.util.Properties;
import javax.swing.JOptionPane;
import math.Calibrator;
import math.Polynomial;

/**
 * Class used to parse raw channel data from a CSV file and generate Channel objects from it.
 * @author Lauritz Møllersen
 */
public class Parser {

    /**
     * Method for parsing a CSV file.
     * @param filename The directory and filename of the CSV file.
     * @param prop 
     * @return ArrayList of Channel objects.
     */
    public static ArrayList<Channel> parseCSV(String filename, Properties prop) {
        ArrayList<Channel> channelArray = null;

        try {
            int amountOfLines = -1;

            // find amount of lines
            InputStream is = new BufferedInputStream(new FileInputStream(filename));
            try {
                byte[] c = new byte[1024];
                int readChars = 0;
                while ((readChars = is.read(c)) != -1) {
                    for (int i = 0; i < readChars; ++i) {
                        if (c[i] == '\n') {
                            ++amountOfLines;
                        }
                    }
                }

            } catch (IOException e) {
                ErrorLogger.addError(new Error(e, 1, "File reading error", "Error reading the CSV file"));
            } finally {
                is.close();
            }

            //create BufferedReader to read the csv file
            BufferedReader br = new BufferedReader(new FileReader(filename));
            String strLine = "";
            channelArray = new ArrayList<Channel>();

            int amountOfTokens = 0;

            // reads the names of the channels from csv file
            while ((strLine = br.readLine()) != null) {
                int i = 0;
                int j = strLine.indexOf(',');
                while (j >= 0) {
                    try {
                        channelArray.add(new Channel(strLine.substring(i, j).trim()));
                    } catch (IndexOutOfBoundsException e) {
                        ErrorLogger.addError(new Error(e, 1, "Format error", "There was a problem with the formatting of your CSV file, line #1"));
                    } finally {
                        amountOfTokens++;
                        i = j + 1;
                        j = strLine.indexOf(',', i);
                    }
                }
                try {
                    channelArray.add(new Channel(strLine.substring(i)));
                    amountOfTokens++;
                } catch (IndexOutOfBoundsException e) {
                    ErrorLogger.addError(new Error(e, 1, "Format error", "There was a problem with the formatting of your CSV file, line #1"));
                }
                break;
            }

            // create the double array for storing channel values.
            double[][] doubleArray = new double[amountOfTokens][amountOfLines];

            //read comma separated file line by line
            String substring = "";

            int tokenCount = 0;
            int lineCount = 0;
            while ((strLine = br.readLine()) != null && lineCount < amountOfLines) {
                tokenCount = 0;
                int i = 0;
                int j = strLine.indexOf(',');

                while (j >= 0) {
                    try {
                        substring = strLine.substring(i, j);
                        /* If the entry is blank, that means the hz is different for the
                         * different channels in the CSV file. Fill that blank with the MAX_VALUE
                         * of a double, then use the interpolate method further down
                         */
                        if (substring.charAt(0) == '?') {
                            doubleArray[tokenCount][lineCount] = Double.MAX_VALUE;
                        } else {
                            doubleArray[tokenCount][lineCount] = (Double.parseDouble(substring));
                        }
                        tokenCount++;
                    } catch (NumberFormatException e) {
                        ErrorLogger.addError(new Error(e, 1, "Format error", "There was a problem with the formatting in your CSV file, see line " + lineCount));
                    } finally {
                        i = j + 1;
                        j = strLine.indexOf(',', i);
                    }
                }
                try {
                    doubleArray[tokenCount][lineCount] = (Double.parseDouble(strLine.substring(i)));
                } catch (NumberFormatException e) {
                    ErrorLogger.addError(new Error(e, 1, "Format error", "There was a problem with the formatting in your CSV file, see line " + lineCount));
                } finally {
                    tokenCount++;
                    lineCount++;
                }
            }

            // interpolating between values if they're missing (for instance if one channel is 10hz and the other is 100hz)
            double tempValue = 0.0;
            double x0 = 0.0;
            double[] tempDoubleArray;


            for (int i = 0; i < amountOfTokens; i++) {
                for (int j = 0; j < amountOfLines; j++) {
                    tempValue = doubleArray[i][j];
                    if (tempValue == Double.MAX_VALUE && j > 0) {
                        tempDoubleArray = doubleArray[i];
                        x0 = doubleArray[i][j - 1];
                        i += interpolate(tempDoubleArray, x0, j);
                    }
                }
            }

            // set the channel values and calibrate from Properties-file
            Channel tempChannel;
            String tempChanName = "";
            HashMap hm = new HashMap<String, Integer>();
            hm.put("Distance", Constants.CHAN_DISTANCE);
            hm.put("Time", Constants.CHAN_TIME);
            hm.put("Speed", Constants.CHAN_SPEED);
            hm.put("Throttle", Constants.CHAN_THROTTLE);
            hm.put("Brake", Constants.CHAN_BRAKE);
            hm.put("LateralG", Constants.CHAN_LATACC);
            hm.put("Lap", Constants.CHAN_LAP);
            hm.put("GpsPosX", Constants.GPS_POS_X);
            hm.put("GpsPosY", Constants.GPS_POS_Y);
    
            for (int i = 0; i < doubleArray.length; i++) {
                tempChannel = channelArray.get(i);
                tempChanName = tempChannel.getName();
                if(hm.containsKey(tempChanName)) {
                    tempChannel.setType((Integer) hm.get(tempChanName));
                }
                
                tempChannel.setValues((doubleArray[i]));
                // calibrate from the calibration-file
                if (prop != null && prop.containsKey(tempChanName)) {
                    Polynomial poly = new Polynomial();
                    String tempPoly = prop.getProperty(tempChanName);
                    int j = 0;
                    int k = tempPoly.indexOf('|');
                    while (k >= 0) {
                        try {
                            poly.add(Float.parseFloat(tempPoly.substring(j, k).trim()));
                        } catch (NumberFormatException e) {
                            ErrorLogger.addError(new Error(e, 1, "Format error", 
                                    "There was a problem with the formatting of your properties file under the property " + tempChanName));
                        } finally {
                            j = k + 1;
                            k = tempPoly.indexOf('|', j);
                        }
                    }
                    poly.add(Float.parseFloat(tempPoly.substring(j).trim()));
                    
                    tempChannel.setPolynomials(poly);
                    Calibrator.calibrate(tempChannel);
                }
                
                // set type
                
            }
            JOptionPane.showMessageDialog(null, (amountOfLines+1) + " lines of data read successfully!");

        } catch (FileNotFoundException e) {
            ErrorLogger.addError(new Error(e, 1, "File not found", "The filename you entered was not found"));
        } catch (IOException e) {
            ErrorLogger.addError(new Error(e, 1, "File reading error", "There was a problem with reading the CSV file"));
        } catch (Exception e) {
            ErrorLogger.addError(new Error(e, 1, "Error", "There was a problem with reading the CSV file"));
        } finally {
            return channelArray;
        }
    }

    /**
     * Method to interpolate between values if some are missing (for instance if one of
     * the sensors capture data at 10hz and the other at 100hz).
     * @param values The values to interpolate on.
     * @param x0 The "start-value" for the interpolating.
     * @param j Where to start iterating through the list of values.
     * @return
     */
    public static int interpolate(double[] values, final double x0, final int j) {
        int nextValue = j;
        double x1 = 0.0;
        while (nextValue < (j + 100) && nextValue < values.length) {
            x1 = values[nextValue];
            nextValue++;
            if (x1 != Double.MAX_VALUE) {
                int difference = nextValue - j;
                for (int i = 0; i < difference; i++) {
                    values[j + i] = x0 + (x1 / difference) * (i + 1);
                }
                return difference;
            }
        }
        return 1;
    }
}
