package plugin;

import java.awt.Component;
import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import project.Command;
import project.Project;
import view.View;

/**
 * Class implementing the default plugin template.
 * @author Per Øyvind Stadheim
 */
public class DefaultPlugin implements Plugin, Externalizable {
    private static final long serialVersionUID = 1L;
    private View parentView;  
    private String title;
    private int cursorPosition;
    private boolean cursorLinked = true; 

    /**
     * Initializes the plugin by setting view reference and plugin ID
     * @param v The parent view containing the plugin
     * @throws PluginInitException  
     */
    @Override
    public void init(View v) throws PluginInitException {
        this.parentView = v;
        if(cursorLinked)this.cursorPosition = v.getGlobalCursorPosition();
        
    }
   
 /**
     * Sets the title of the plugin. This method is only used by the
     * PluginFactory for initialization purposes.
     * @param title
     */
    @Override
    public void setTitle(String title) {
        this.title = title;
    }
    
       /**
     * Get a reference to the parent view.
     * @return The parent view.
     */
    @Override   
    public View getParentView() {
        return this.parentView;
    }
          
/**
     * Sets the plugin title
     * @param title Plugin title
     * @param undo
     */
    @Override
    public void setTitle(String title, boolean undo) {
         if(!undo){
            this.title = title;
            //Notify view
            getParentView().pluginTitlesChanged();
       } else {
            if(!title.equals(this.getTitle())){
            final String oldValue = this.getTitle();
            final String newValue = title;
            final DefaultPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setTitle(newValue, false);
                }

                @Override
                public void reverse() {
                    source.setTitle(oldValue, false);
                }

                @Override
                public String toString() {
                    return "Set Plugin Title";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
       }
           
         }
        
    }
    
    /**
     * Returns the current plugin title
     * @return Plugin title
     */
    @Override
    public String getTitle() {
        return title;
    }  
     
   /**
     * Set whether the cursor shall be linked with the parent view or not.
     * @param b Value
     * @param undo
     */
    @Override
    public void setCursorLinked(boolean cursorLinked, boolean undo) {
       if(!undo){
            this.cursorLinked = cursorLinked;
            this.cursorPosition = getParentView().getGlobalCursorPosition();
       } else {
            final boolean oldValue = isCursorLinked();
            final boolean newValue = cursorLinked;
            final DefaultPlugin source = this; 
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setCursorLinked(newValue, false);
                }

                @Override
                public void reverse() {
                    source.setCursorLinked(oldValue, false);
                }

                @Override
                public String toString() {
                    return "Set Link Cursor";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
       }
    }
        
   /**
     * Returns if the cursor is linked with its parent view or not.
     * @return
     */
    @Override
    public boolean isCursorLinked() {
        return cursorLinked;
    }
    
    /**
     * Sets the plugin's cursor position
     * @param position Cursor position
     */
    @Override
    public void setCursorPosition(int cursorPosition) {
        this.cursorPosition = cursorPosition;
    }   
    
    /**
     * Returns the current cursor position
     * @return Cursor position
     */
    @Override
    public int getCursorPosition() {
        return cursorPosition;
    }
    
    /**
     * Method used for plugins needing to prepare for playback
     */
    @Override
    public void startPlayback() { 
    }
    
    /**
     * Method used for stopping playback for plugins who need that.
     */
    @Override
    public void stopPlayback() {
    }
    
    
    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeObject(this.title);
        out.writeBoolean(this.cursorLinked);
        out.writeInt(this.cursorPosition);
    }

    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        this.title = (String)in.readObject();
        this.cursorLinked = in.readBoolean();
        this.cursorPosition = in.readInt();
    }
       
    /**
     * Method used to update the plugin based on changes in the 
     * segmentation.
     */
    @Override
    public void segmentationChanged() { 
    }

    /**
     * Method used to update the plugin based on changes in the 
     * data channels.
     */
    @Override
    public void channelsChanged() {
    }

    /**
     * Method used to update the plugin based on changes in the
     * global lap or channel coloring scheme.
     */
    @Override
    public void colorsChanged() {
    }
    
    /**
     * Get the main GUI component of the Plugin.
     * @return GUI component.
     */
    @Override
    public Component getPluginPanel() {
        throw new UnsupportedOperationException("Not supported yet.");
    }
    /**
     * Get the property GUI component of the Plugin. 
     * @return GU /**I component.
     */
    @Override
    public Component getPropertiesPanel() {
        throw new UnsupportedOperationException("Not supported yet.");
    }
    
    /**
     * 
     * @return
     */
    public Project getParentProject() {
        return this.parentView.getParentProject();
    }   
}