package plugin;

import java.awt.Component;
import segmentation.Segmentation;
import view.View;


/**
 * The Plugin interface.
 * @author Per Øyvind Stadheim
 */
public interface Plugin {

   /**
     * Initializes the plugin by setting view reference and plugin ID
     * @param v The parent view containing the plugin
     * @throws PluginInitException  
     */
    public void init(View v) throws PluginInitException;
   
    /**
     * Sets the title of the plugin. This method is only used by the
     * PluginFactory for initialization purposes.
     * @param title
     */
    public void setTitle(String title);
    
   /**
     * Get a reference to the parent view.
     * @return The parent view.
     */
    public View getParentView();
    
    /**
     * Sets the plugin title
     * @param title Plugin title
     * @param undo
     */
    public void setTitle(String title, boolean undo);
    
    /**
     * Returns the current plugin title
     * @return Plugin title
     */
    public String getTitle();
        
    /**
     * Set whether the cursor shall be linked with the parent view or not.
     * @param b Value
     * @param undo
     */
    public void setCursorLinked(boolean b, boolean undo);
    
    /**
     * Returns if the cursor is linked with its parent view or not.
     * @return
     */
    public boolean isCursorLinked();
        
    /**
     * Sets the plugin's cursor position
     * @param position Cursor position
     */
    public void setCursorPosition(int position);
    
    /**
     * Returns the current cursor position
     * @return Cursor position
     */
    public int getCursorPosition();
    
    /**
     * Method used for plugins needing to prepare for playback
     */
    public void startPlayback();
    
    /**
     * Method used for stopping playback for plugins who need that.
     */
    public void stopPlayback();
    
    /**
     * Method used to update the plugin based on changes in the 
     * segmentation.
     */
    public void segmentationChanged();
    
    /**
     * Method used to update the plugin based on changes in the 
     * data channels.
     */
    public void channelsChanged();

    /**
     * Method used to update the plugin based on changes in the
     * global lap or channel coloring scheme.
     */
    public void colorsChanged();
    
    /**
     * Get the main GUI component of the Plugin.
     * @return GUI component.
     */
    public Component getPluginPanel();

    /**
     * Get the property GUI component of the Plugin. 
     * @return GU /**I component.
     */
    public Component getPropertiesPanel();
    
    
}
