package plugin;

import error.ErrorLogger;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import serializer.Serializer;

/**
 * Class used to instantiate Plugin objects by Java reflection.
 * @author Per Øyvind Stadheim
 */
public class PluginFactory {

    /**
     * Loads plugins from jars and presents them in a list of class objects.
     * @return List of available plugin classes
     */
    public static ArrayList<Class> loadPluginsFromJars() {
        ArrayList<Class> classes = new ArrayList<Class>();
        ArrayList<File> files = PluginDependencyLoader.getJarFiles();
        for (File f : files) {
            if (f.getName().contains("plugin")) {
                try {
                    String clazz = f.getName().substring(0, f.getName().length() - 4);//"plugin.testplugin.TestPlugin";
                    URL u;
                    try {
                        u = f.toURI().toURL();
                        URLClassLoader cl = new URLClassLoader(new URL[]{u});
                        try {
                            Object o = Class.forName(clazz, true, cl).newInstance();
                            if (o instanceof Plugin) {
                                classes.add(o.getClass());
                            }
                        } catch (InstantiationException ex) {
                        } catch (IllegalAccessException ex) {
                        }
                    } catch (MalformedURLException ex) {
                    }
                } catch (ClassNotFoundException ex) {
                }
            }
        }
        return classes;
    }

    /**
     * Creates a new plugin instance, based on classpath.
     * @param type The classpath of the plugin to be instantiated.
     * @param counter 
     * @return The new plugin instance.
     */
    public static Plugin createPlugin(Class type, int counter) { // add plugincounter
        try {
            Plugin newPlugin = (Plugin) type.newInstance();
            String fullTypeString = type.getName();
            int index = fullTypeString.lastIndexOf(".");
            String typeString = fullTypeString.substring(index + 1);
            newPlugin.setTitle("Plugin " + counter + " (" + typeString + ")");
            return newPlugin;
        } catch (InstantiationException ex) {
        } catch (IllegalAccessException ex) {
        }

        return null;
    }

    /**
     * Serializes a plugin to a ByteArray
     * @param p Input plugin
     * @return ByteArray representation of the plugin
     */
    public static byte[] serializeToByteArray(Plugin p) {
        return Serializer.serializeToByteArray(p);
    }

    /**
     * Deserializes a plugin from a ByteArray
     * @param b Input ByteArray
     * @return Plugin
     * @throws IOException
     */
    public static Plugin deserializeFromByteArray(byte[] b) throws IOException {
        try {
            return (Plugin) Serializer.deserializeFromByteArray(b);
        } catch (ClassNotFoundException ex) {
            ErrorLogger.addError(new error.Error("Plugin initialization error", "A plugin component is missing from the system."));
        }
        return null;
    }
}
