package plugin.linechart;

import error.ErrorLogger;
import plugin.Plugin;
import java.awt.Component;
import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.ArrayList;
import javax.swing.JPanel;
import plugin.DefaultPlugin;
import plugin.PluginInitException;
import project.Command;
import view.View;
import parser.Constants;

/**
 * Class that implements the linechart plugin prototype.
 * @author Per Øyvind Stadheim
 */
public class LineChartPlugin extends DefaultPlugin {

    private static final long serialVersionUID = 1L;
    private int dataVersion = 1;
    private int xAxis = Constants.CHAN_DISTANCE;
    private ArrayList<String> selectedChannels;
    private ArrayList<Integer> selectedLaps;
    private ArrayList<Integer> selectedSegments;
    private boolean lapsSuperImposed = true;
    private boolean followDomainCursor = true;
    private PluginPanel pluginPanel;
    private TabbedPropPanel propPanel;

    /**
     * 
     */
    public LineChartPlugin() {
    }

    /**
     * Initializes the plugin by setting view reference and plugin ID. This override
     * also contains specific initialization code for this plugin.
     * @param v The parent view of the plugin
     * @throws PluginInitException  
     */
    @Override
    public void init(View v) throws PluginInitException {
        super.init(v);
        //Check if dataset supports plugin
        if (getParentProject().getChannelType(Constants.CHAN_DISTANCE) == null
                || getParentProject().getChannelType(Constants.CHAN_TIME) == null) {
            ErrorLogger.addError(new error.Error("Channel error", "Error with the time and distance channels. "
                    + "Check that you have defined these in the Channel Options."));
            throw new PluginInitException();
        }


        //If not loaded from file, intialize default fields.
        if (selectedChannels == null) {
            selectedChannels = new ArrayList<String>();
        }
        if (selectedLaps == null) {
            selectedLaps = new ArrayList<Integer>();
        }
        if (selectedSegments == null) {
            selectedSegments = new ArrayList<Integer>();
        }

        //Check and eventually nerf plugin settings vs current project (segmentation, channels)
        //  removeUnsupportedChannelValues();
        //removeUnsupportedLapValues();
        //removeUnsupportedSegmentValues();

        //Create GUI components
        propPanel = new TabbedPropPanel();
        propPanel.setLcp(this);
        pluginPanel = new PluginPanel(this);
        propPanel.setup();
        pluginPanel.generateDomainSegmentOffsets();
        if (!selectedChannels.isEmpty()) {
            pluginPanel.rebuildSeries();
        }
    }

    // Methods used internally
    private void removeUnsupportedLapValues() {
        for (int i = 0; i < selectedLaps.size(); i++) {
            boolean exists = false;
            if (selectedLaps.get(i) <= getParentProject().getSegmentation().getAmountOfLaps()) {
                exists = true;
            }
            if (!exists) {
                selectedLaps.remove(i);
                i--;
            }
        }
    }

    private void removeUnsupportedSegmentValues() {
        for (int i = 0; i < selectedSegments.size(); i++) {
            boolean exists = false;
            if (selectedSegments.get(i) <= getParentProject().getSegmentation().getAmountOfSegments() + 1) {
                exists = true;
            }
            if (!exists) {
                selectedSegments.remove(i);
                i--;
            }
        }
    }

    private void removeUnsupportedChannelValues() {
        for (int i = 0; i < selectedChannels.size(); i++) {
            boolean exists = false;
            String s1 = selectedChannels.get(i);
            for (String s2 : getParentProject().getChanStrings()) {
                if (s1.equals(s2)) {
                    exists = true;
                    break;
                }
                for (String s3 : getParentProject().getMathChanStrings()) {
                    if (s1.equals(s3)) {
                        exists = true;
                        break;
                    }
                    if (!exists) {
                        selectedChannels.remove(i);
                        i--;
                    }
                }
            }
        }
    }

    /**
     * 
     * @param b
     * @param undo
     */
    @Override
    public void setCursorLinked(boolean b, boolean undo) {
        super.setCursorLinked(b, undo);
        propPanel.linkCursorChanged();
    }

    /**
     * 
     * @param channel
     * @param value
     * @param undo
     */
    public void setChannelSelection(String channel, boolean value, boolean undo) {
        if (!undo) {
            if (!value && selectedChannels.contains(channel)) {
                selectedChannels.remove(channel);
            } else if (value && !selectedChannels.contains(channel)) {
                selectedChannels.add(channel);
            }
            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.channelsChanged();
        } else {
            final String targetChannel = channel;
            final boolean oldValue = selectedChannels.contains(channel);
            final boolean newValue = value;
            final LineChartPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setChannelSelection(targetChannel, newValue, false);
                }

                @Override
                public void reverse() {
                    source.setChannelSelection(targetChannel, oldValue, false);
                }

                @Override
                public String toString() {
                    return "Channel Selection";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public ArrayList<String> getSelectedChannels() {
        return this.selectedChannels;
    }

    /**
     * 
     * @param lap
     * @param value
     * @param undo
     */
    public void setLapSelection(int lap, boolean value, boolean undo) {
        if (!undo) {
            if (!value && selectedLaps.contains(new Integer(lap))) {
                selectedLaps.remove(new Integer(lap));
            } else if (value && !selectedLaps.contains(lap)) {
                selectedLaps.add(lap);
            }

            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.lapsChanged();
        } else {
            final int targetLap = lap;
            final boolean oldValue = selectedLaps.contains(lap);
            final boolean newValue = value;
            final LineChartPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setLapSelection(targetLap, newValue, false);
                }

                @Override
                public void reverse() {
                    source.setLapSelection(targetLap, oldValue, false);
                }

                @Override
                public String toString() {
                    return "Lap Selection";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public ArrayList<Integer> getSelectedLaps() {
        return this.selectedLaps;
    }

    /**
     * 
     * @param segment
     * @param value
     * @param undo
     */
    public void setSegmentSelection(int segment, boolean value, boolean undo) {
        if (!undo) {
            if (!value && selectedSegments.contains(new Integer(segment))) {
                selectedSegments.remove(new Integer(segment));
            } else if (value && !selectedSegments.contains(segment)) {
                selectedSegments.add(segment);
            }
            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.segmentsChanged();
        } else {
            final int targetSegment = segment;
            final boolean oldValue = selectedSegments.contains(segment);
            final boolean newValue = value;
            final LineChartPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setSegmentSelection(targetSegment, newValue, false);
                }

                @Override
                public void reverse() {
                    source.setSegmentSelection(targetSegment, oldValue, false);
                }

                @Override
                public String toString() {
                    return "Segment Selection";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public ArrayList<Integer> getSelectedSegments() {
        return this.selectedSegments;
    }

    /**
     * 
     * @param superImp
     * @param undo
     */
    public void setLapsSuperImposed(boolean superImp, boolean undo) {
        if (!undo) {
            this.lapsSuperImposed = superImp;
            //Updating GUI
            pluginPanel.rebuildSeries();
            propPanel.lapModeChanged();
        } else {
            final boolean oldValue = this.isLapsSuperImposed();
            final boolean newValue = superImp;
            final LineChartPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setLapsSuperImposed(newValue, false);
                }

                @Override
                public void reverse() {
                    source.setLapsSuperImposed(oldValue, false);
                }

                @Override
                public String toString() {
                    return "Change Lap Mode";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public boolean isLapsSuperImposed() {
        return this.lapsSuperImposed;
    }

    /**
     * 
     * @param b
     * @param undo
     */
    public void setFollowDomainCursor(boolean b, boolean undo) {
        if (!undo) {
            this.followDomainCursor = b;
            //Update GUI
            pluginPanel.setFollowDomainCursor(b);
            propPanel.followCursorChanged();
        } else {
            final boolean oldValue = this.isFollowDomainCursor();
            final boolean newValue = b;
            final LineChartPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setFollowDomainCursor(newValue, false);
                }

                @Override
                public void reverse() {
                    source.setFollowDomainCursor(oldValue, false);
                }

                @Override
                public String toString() {
                    return "Set Follow Cursor";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public boolean isFollowDomainCursor() {
        return followDomainCursor;
    }

    /**
     * 
     * @param axis
     * @param undo
     */
    public void setXAxis(int axis, boolean undo) {
        if (!undo) {
            this.xAxis = axis;
            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.xAxisChanged();
        } else {
            final int oldValue = this.getXAxis();
            final int newValue = axis;
            final LineChartPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setXAxis(newValue, false);
                }

                @Override
                public void reverse() {
                    source.setXAxis(oldValue, false);
                }

                @Override
                public String toString() {
                    return "Change X-Axis";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public int getXAxis() {
        return this.xAxis;
    }

    /**
     * 
     * @param title
     * @param undo
     */
    @Override
    public void setTitle(String title, boolean undo) {
        super.setTitle(title, undo);

        //Update GUI
        if (propPanel != null) {
            propPanel.titleChanged();
        }

    }

    // Methods used externally
    /**
     * 
     */
    @Override
    public void channelsChanged() {
        //removeUnsupportedChannelValues();
        pluginPanel.rebuildSeries();
        propPanel.channelsChanged();

    }

    /**
     * 
     */
    @Override
    public void colorsChanged() {
        pluginPanel.rebuildSeries();
        propPanel.channelColorsChanged();

        propPanel.lapColorsChanged();
    }

    /**
     * 
     */
    @Override
    public void segmentationChanged() {
        //removeUnsupportedSegmentValues();
        pluginPanel.rebuildSeries();
        pluginPanel.generateDomainSegmentOffsets();
        propPanel.lapsChanged();
        propPanel.segmentsChanged();
    }

    /**
     * 
     * @param position
     */
    @Override
    public synchronized void setCursorPosition(int position) {
        super.setCursorPosition(position);
        //Updating GUI
        pluginPanel.setCursorPosition(position);
    }

    /**
     * Get the main GUI component of the Plugin.
     * @return GUI component.
     */
    @Override
    public JPanel getPluginPanel() {
        return this.pluginPanel;

    }

    /**
     * Get the property GUI component of the Plugin. 
     * @return GU /**I component.
     */
    @Override
    public Component getPropertiesPanel() {
        return this.propPanel;
    }

    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        removeUnsupportedChannelValues();
        removeUnsupportedLapValues();
        removeUnsupportedSegmentValues();
        super.writeExternal(out);
        out.writeInt(this.dataVersion);
        out.writeObject(this.getSelectedChannels());
        out.writeObject(this.getSelectedLaps());
        out.writeObject(this.getSelectedSegments());
        out.writeBoolean(lapsSuperImposed);
        out.writeBoolean(followDomainCursor);

    }

    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        super.readExternal(in);
        int version = in.readInt();
        if (version == 1) {
            this.selectedChannels = (ArrayList<String>) in.readObject();
            this.selectedLaps = (ArrayList<Integer>) in.readObject();
            this.selectedSegments = (ArrayList<Integer>) in.readObject();
            this.lapsSuperImposed = in.readBoolean();
            this.followDomainCursor = in.readBoolean();

        }
    }
}
