/**
 * Class that implements the track map plugin prototype.
 * @author Lauritz Møllersen
 */
package plugin.map;

import math.Map;
import channel.Channel;
import java.awt.Component;
import java.awt.Color;
import java.awt.RenderingHints;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Arrays;


import math.MathFunction;
import org.jfree.chart.ChartMouseEvent;
import org.jfree.chart.ChartMouseListener;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYDotRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.RectangleInsets;
import plugin.DefaultPlugin;
import project.Project;
import segmentation.Segmentation;
import view.View;
import error.Error;
import error.ErrorLogger;
import java.awt.Point;
import java.util.Collections;
import javax.swing.JOptionPane;
import org.jfree.chart.ChartRenderingInfo;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.ui.RectangleEdge;
import plugin.PluginInitException;
import parser.Constants;
import project.Command;

/**
 *
 * @author Per Øyvind
 */
public class MapPlugin extends DefaultPlugin {

    private static final long serialVersionUID = 1L;
    private ChartPanel chartPanel;
    private XYSeriesCollection collection1;
    private XYSeriesCollection collection2;
    private Channel throttle;
    private Channel gpsx;
    private Channel gpsy;
    private Channel speed;
    private Channel latacc;
    private double[] distanceValues;
    private Map map;
    private XYSeries dataBrake = new XYSeries("Throttle < 30%");
    private XYSeries dataThrottle = new XYSeries("Throttle > 85%");
    private XYSeries dataNeutral = new XYSeries("30-85% throttle");
    private XYSeries dataStartPoint = new XYSeries("Start");
    private XYSeries segmentPoints = new XYSeries("Segment");
    private Project p;
    private Segmentation s;
    private int startOffset;
    private int endOffset;
    private MapPropPanel propPanel;
    private int size;
    private double[] x;
    private double[] y;
    private XYItem[] xyitems;
    private ArrayList<Integer> segmentOffsets;
    private boolean createSegments = false;
    private int lap = -1;
    private boolean gps;

    /**
     * 
     */
    public MapPlugin() {
    }

    /**
     * Initializes the plugin by setting view reference and plugin ID. This override
     * also contains specific initialization code for this plugin.
     * @param v The parent view of the plugin
     * @throws PluginInitException  
     */
    @Override
    public void init(View v) throws PluginInitException {
        super.init(v);
        p = v.getParentProject();
        gps = p.isGps();
        throttle = p.getChannelType(Constants.CHAN_THROTTLE);
        gpsx = p.getChannelType(Constants.GPS_POS_X);
        gpsy = p.getChannelType(Constants.GPS_POS_Y);
        speed = p.getChannelType(Constants.CHAN_SPEED);
        latacc = p.getChannelType(Constants.CHAN_LATACC);
        Channel tempDistance = p.getChannelType(Constants.CHAN_DISTANCE);

        if (throttle == null || tempDistance == null) {
            ErrorLogger.addError(new Error("Channel error", "Error with the brake, throttle or distance channels. "
                    + "Check that you have defined these in the Channel Options."));
            throw new PluginInitException();
        } else if (gps == true && (gpsx == null || gpsy == null)) {
            ErrorLogger.addError(new Error("Channel error", "GPS is set to enabled, but there is an error with the GPS channels. "
                    + "Check that you have defined these in the Channel Options."));
            throw new PluginInitException();
        } else if (gps == false && (speed == null || latacc == null)) {
            ErrorLogger.addError(new Error("Channel error", "GPS is set to enabled, but there is an error with the speed and lateral acceleration channels. "
                    + "Check that you have defined these in the Channel Options."));
            throw new PluginInitException();
        }



        Segmentation tempSeg = p.getSegmentation();
        if (tempSeg == null) {
            s = new Segmentation();
            s.createLaps(p.getChannelType(Constants.CHAN_LAP));
            p.setSegmentation(s);
        } else {
            s = tempSeg;
        }



        propPanel = new MapPropPanel(this);

        distanceValues = p.getChannelType(Constants.CHAN_DISTANCE).getCalibratedValues();

        JFreeChart chart = createChart();
        createChartPanel(chart);

        if (isCursorLinked()) {
            setCursorPosition(v.getGlobalCursorPosition());
        } else {
            setCursorPosition(0);
        }

        xyitems = doubleToXYItem(x, y);
        Arrays.sort(xyitems);

        setSegmentStrings();
        setCursorLinked(true, false);

    }

    /**
     * 
     * @param linked
     * @param undo
     */
    public void setCursorLinked(boolean linked, boolean undo) {
        if (!undo) {
            super.setCursorLinked(linked, undo);
            //Update GUI
            propPanel.updateSyncCheckBox();
        } else {
            final boolean newValue = linked;
            final boolean oldValue = isCursorLinked();
            final MapPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setCursorLinked(newValue, false);
                }

                @Override
                public void reverse() {
                    source.setCursorLinked(oldValue, false);
                }

                @Override
                public String toString() {
                    return "Set Link Cursor";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);

        }
    }

    /**
     * Get the main GUI component of the Plugin.
     * @return GUI component.
     */
    @Override
    public ChartPanel getPluginPanel() {
        return this.chartPanel;
    }

    private void createChartPanel(JFreeChart chart) {
        chartPanel = new ChartPanel(chart);

        chartPanel.setMouseWheelEnabled(false);
        chartPanel.setDomainZoomable(true);
        chartPanel.setRangeZoomable(true);
        chartPanel.setMouseWheelEnabled(true);
        chartPanel.getChart().setAntiAlias(false);
        chartPanel.removeChartMouseListener(null);

        chartPanel.addChartMouseListener(new ChartMouseListener() {

            @Override
            public void chartMouseClicked(ChartMouseEvent e) {
                // for creating segments on the map
                JFreeChart chart = (JFreeChart) e.getSource();
                int mouseX = e.getTrigger().getX();
                int mouseY = e.getTrigger().getY();
                Point2D p = chartPanel.translateScreenToJava2D(
                        new Point(mouseX, mouseY));
                XYPlot plot = (XYPlot) chart.getPlot();
                ChartRenderingInfo info = chartPanel.getChartRenderingInfo();
                Rectangle2D dataArea = info.getPlotInfo().getDataArea();

                ValueAxis domainAxis = plot.getDomainAxis();
                RectangleEdge domainAxisEdge = plot.getDomainAxisEdge();
                ValueAxis rangeAxis = plot.getRangeAxis();
                RectangleEdge rangeAxisEdge = plot.getRangeAxisEdge();
                double tempX = domainAxis.java2DToValue(p.getX(), dataArea,
                        domainAxisEdge);
                double tempY = rangeAxis.java2DToValue(p.getY(), dataArea,
                        rangeAxisEdge);

                int offset = quickfind(tempX, tempY);

                if (createSegments == true) {
                    if (offset >= 0) {
                        segmentOffsets.add(offset);
                        Collections.sort(segmentOffsets);
                        final int offsetNumber = segmentOffsets.indexOf(offset);
                        final int relativeOffset = offset - startOffset;

                        XYSeries temp = collection1.getSeries(0);

                        temp.add(x[relativeOffset], y[relativeOffset]);

                        propPanel.addSegmentString("S" + (offsetNumber + 1) + ": " + MathFunction.roundTwoDecimals(distanceValues[startOffset + offset] - distanceValues[startOffset]) + "m.", offsetNumber);

                    }
                } else {
                    if (isCursorLinked()) {
                        offset = slowfind(tempX, tempY, x, y);
                        getParentView().setGlobalCursorPosition(s.getLapOffset(lap) + offset);
                    }
                }
            }

            @Override
            public void chartMouseMoved(ChartMouseEvent e) {
                //
            }
        });
    }

    private int slowfind(double tempX, double tempY, double[] xValues, double[] yValues) {
        if (xValues == null || xValues.length == 0 || yValues == null || yValues.length == 0) {
            return -1;
        }

        double diff = Double.POSITIVE_INFINITY;
        int closestIndex = -1;
        for (int i = 0; i < xValues.length; i++) {
            double tempDiff = Math.abs(Math.abs(tempX) - Math.abs(xValues[i])) + Math.abs(Math.abs(tempY) - Math.abs(yValues[i]));
            if (tempDiff < diff) {


                diff = tempDiff;
                closestIndex = i;

            }
        }

        return closestIndex;
    }

    /**
     * Used in order to create segments from the offset-list.
     */
    public void createSegments() {
        if (segmentOffsets.size() > 0) {
            try {
                s.createSegments(segmentOffsets, (Integer) getLap(), p.getChannelType(1));
                getParentProject().setSegmentation(s);
                JOptionPane.showMessageDialog(null, segmentOffsets.size() + " segments created successfully!");
            } catch (Exception e) {
                error.ErrorLogger.addError(new Error(e, 1, "Error creating segments", "Check segments and raw data"));
            }
        }
    }

    /**
     * Used for simulation, this method is called by the PlayBack plugin.
     * @param absoluteposition 
     */
    @Override
    public void setCursorPosition(int absoluteposition) {
        int position = absoluteposition - s.getLapOffset(s.getLapFromOffset(absoluteposition));
        int newLap = s.getLapFromOffset(absoluteposition);
        if (newLap == lap) {
            XYPlot plot = chartPanel.getChart().getXYPlot();
            final double xValue = x[position];
            final double yValue = y[position];
            plot.setDomainCrosshairValue(xValue);
            plot.setRangeCrosshairValue(yValue);
            // plot.getDomainAxis().centerRange(  plot.getDomainCrosshairValue());
            // plot.getRangeAxis().centerRange(plot.getRangeCrosshairValue());
        } else if (newLap == -1) {
            ErrorLogger.addError(new Error("Invalid lap number", "The position recieved from the Playback plugin does not match a valid lap"));
        } else {
            setLap(newLap);
            XYPlot plot = chartPanel.getChart().getXYPlot();
            final double xValue = x[position];
            final double yValue = y[position];
            plot.setDomainCrosshairValue(xValue);
            plot.setRangeCrosshairValue(yValue);
            // plot.getDomainAxis().centerRange(  plot.getDomainCrosshairValue());
            // plot.getRangeAxis().centerRange(plot.getRangeCrosshairValue());
        }

    }

    private void setData() {
        dataBrake = new XYSeries(">30% throttle");
        dataThrottle = new XYSeries("<85% throttle");
        dataNeutral = new XYSeries("30-85% throttle");
        dataStartPoint = new XYSeries("Start");
        segmentPoints = new XYSeries("Segments");
        size = endOffset - startOffset;
        double tempThrottle = 0.0;
        double[] throttleValues = throttle.getValues();
        dataStartPoint.add(x[0], y[0]);
        for (int i = 1; i < size; i++) {
            tempThrottle = throttleValues[i + startOffset];
            if (tempThrottle > 0.85) {
                dataThrottle.add(x[i], y[i]);
            } else if (tempThrottle < 0.30) {
                dataBrake.add(x[i], y[i]);
            } else {
                dataNeutral.add(x[i], y[i]);
            }
        }
        for (int i = 0; i < segmentOffsets.size(); i++) {
            final int relativeOffset = segmentOffsets.get(i) - startOffset;
            segmentPoints.add(x[relativeOffset], y[relativeOffset]);
        }
    }

    private XYItem[] doubleToXYItem(double[] x, double[] y) {
        XYItem[] xyitems = new XYItem[size];
        for (int i = 0; i < size; i++) {
            xyitems[i] = new XYItem(x[i], y[i], i);
        }
        return xyitems;
    }

    private int quickfind(final double x, final double y) {
        int i = 0;
        while (xyitems[i].getX() < x - 5.0) {
            i++;
        }
        double smallestDifference = 10;
        int index = -1;
        double tempDiff = 0.0;
        XYItem xyitem;
        while ((xyitem = xyitems[i]).getX() < x + 5.0) {
            tempDiff = Math.abs(x - xyitem.getX()) + Math.abs(y - xyitem.getY());
            if (tempDiff < smallestDifference) {
                index = i;
                smallestDifference = tempDiff;
            }
            i++;
        }
        if (index == -1) {
            return -1;
        } else {
            return xyitems[index].getIndex();
        }
    }

    private JFreeChart createChart() {
        collection1 = new XYSeriesCollection();
        collection2 = new XYSeriesCollection();

        XYPlot plot = new XYPlot();

        XYDotRenderer renderer1 = new XYDotRenderer();
        XYDotRenderer renderer2 = new XYDotRenderer();
        renderer1.setSeriesPaint(0, new Color(0, 0, 0));
        renderer1.setSeriesPaint(1, new Color(255, 255, 0));
        renderer2.setSeriesPaint(0, new Color(220, 0, 0));
        renderer2.setSeriesPaint(1, new Color(60, 60, 255));
        renderer2.setSeriesPaint(2, new Color(0, 100, 0));

        renderer1.setDotHeight(12);
        renderer1.setDotWidth(12);

        renderer2.setDotHeight(6);
        renderer2.setDotWidth(6);

        ValueAxis domain = new NumberAxis("");
        ValueAxis range = new NumberAxis("");

        domain.setVisible(false);
        range.setVisible(false);

        plot.setDataset(0, collection1);
        plot.setDataset(1, collection2);

        plot.setRenderer(0, renderer1);
        plot.setRenderer(1, renderer2);
        plot.setDomainCrosshairLockedOnData(false);
        plot.setRangeCrosshairLockedOnData(false);
        plot.setDomainAxis(domain);
        plot.setRangeAxis(range);
        plot.setDomainCrosshairVisible(true);
        plot.setRangeCrosshairVisible(true);
        plot.setDomainGridlinesVisible(false);
        plot.setRangeGridlinesVisible(false);
        collection1.addSeries(segmentPoints);
        collection1.addSeries(dataStartPoint);
        collection2.addSeries(dataBrake);
        collection2.addSeries(dataNeutral);
        collection2.addSeries(dataThrottle);
        plot.setBackgroundPaint(Color.WHITE);
        plot.setAxisOffset(RectangleInsets.ZERO_INSETS);
        plot.setInsets(RectangleInsets.ZERO_INSETS);
        //plot.setDomainAxis(0, null);
        //plot.setRangeAxis(0, null);

        JFreeChart chart = new JFreeChart("Multi Dataset Chart", JFreeChart.DEFAULT_TITLE_FONT, plot, true);

        chart.setPadding(RectangleInsets.ZERO_INSETS);
        chart.setTitle("");
        chart.setBorderVisible(false);

        chart.getRenderingHints().put(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_PURE);

        return chart;
    }

    /**
     * Get the property GUI component of the Plugin. 
     * @return GU /**I component.
     */
    public Component getPropertiesPanel() {
        return this.propPanel;
    }

    /**
     * 
     */
    public void deleteSegments() {
        segmentOffsets.clear();
        segmentPoints.clear();
        s.deleteSegments();
    }

    /**
     * Sets a property in the plugin with undo functionality.
     * This override also takes care of updating the GUI components when a 
     * property is changed.
     * @param title 
     * @param undo Boolean that enables or disables undo functionality.
     */
    public void setTitle(String title, boolean undo) {
        super.setTitle(title, undo);

        //Update GUI
        if (propPanel != null) {
            propPanel.updateTitle();
        }

    }

    /**
     * @param value The lap you want to change to.
     */
    public void setLap(int value) {
        this.lap = value;
        segmentOffsets = s.getLapSegments(lap);

        startOffset = s.getLapOffset(lap);
        endOffset = s.getLapEndOffset(lap);


        if (gps) {
            map = new Map(p.getChannelType(Constants.GPS_POS_X).getCalibratedValues(),
                    p.getChannelType(Constants.GPS_POS_Y).getCalibratedValues());
        } else {
            map = MathFunction.calculateMap(p.getChannelType(Constants.CHAN_LATACC).getCalibratedValues(),
                    p.getChannelType(Constants.CHAN_SPEED).getCalibratedValues(),
                    p.getHz(), 0.277777777777778, startOffset, endOffset);
        }

        x = map.getPositionX();
        y = map.getPositionY();

        setData();
        chartPanel.setChart(createChart());
        propPanel.updateLapSelectionBox();
    }

    /**
     * 
     * @return
     */
    public int getLap() {
        return this.lap;
    }

    /**
     * @param createSegments the createSegments to set
     */
    public void setCreateSegments(boolean createSegments) {
        this.createSegments = createSegments;
    }

    /**
     * Updates the strings used by the properties panel to show segments.
     */
    public void setSegmentStrings() {
        final int segmentSize = s.getAmountOfSegments();
        int segmentOffset = 0;
        if (segmentSize > 0) {
            propPanel.deleteSegments();
            for (int i = 0; i < segmentSize; i++) {
                segmentOffset = s.getSegmentOffset(1, i + 1);
                if (segmentOffset >= 0) {
                    propPanel.addSegmentString("S" + (i + 1) + ": " + MathFunction.roundTwoDecimals(distanceValues[startOffset + segmentOffset] - distanceValues[startOffset]) + "m.", i);
                }
            }
        }
    }

    /**
     * Is executed if the channels are changed. Refreshes the GUI.
     */
    @Override
    public void channelsChanged() {
        setLap(lap);
    }

    /**
     * Is executed if the segmentation is changed. Refreshes the GUI.
     */
    @Override
    public void segmentationChanged() {
        setLap(lap);
    }
}
