/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package plugin.report;

import channel.Channel;
import error.ErrorLogger;
import java.text.DecimalFormat;
import math.MathFunction;
import segmentation.Segmentation;

/**
 * Class to represent a channel report in the ReportPlugin.
 * @author Lauritz Møllersen
 */
public class Report {

    private Segmentation segmentation;
    private double[] values;
    private Object[][] splitTimes;
    private ErrorLogger el = new ErrorLogger();
    private DecimalFormat df;

    /**
     * Default constructor.
     * @param segmentation The segmentation.
     * @param timeChannel The time channel.
     */
    public Report(Segmentation segmentation, Channel timeChannel) {
        this.segmentation = segmentation;
        this.values = timeChannel.getCalibratedValues();
        calculateSplitTimes();
    }

    /**
     * @return The table-headers for the report.
     */
    public String[] getHeaders() {
        String[] tempString = new String[segmentation.getAmountOfSegments() + 2];
        tempString[0] = "Laps/Segments";
        for (int i = 1; i < tempString.length; i++) {
            tempString[i] = "S#" + i;
        }
        return tempString;
    }

    private void calculateSplitTimes() {
        final int amountOfLaps = segmentation.getAmountOfLaps();
        final int amountOfSegments = segmentation.getAmountOfSegments();
        splitTimes = new Object[amountOfLaps][amountOfSegments + 2];
        int startOffset = 0;
        int endOffset = 0;
        for (int i = 0; i < amountOfLaps; i++) {
            splitTimes[i][0] = "Lap #" + (i + 1);
            for (int j = 1; j < amountOfSegments + 2; j++) {
                startOffset = segmentation.getSegmentOffset(i + 1, j);
                endOffset = segmentation.getSegmentOffset(i + 1, j - 1);
                if (startOffset < 0 || endOffset < 0) {
                    splitTimes[i][j] = Double.NaN;
                } else {
                    splitTimes[i][j] = MathFunction.roundTwoDecimals((values[startOffset] - values[endOffset]));
                }
            }
        }
    }

    /**
     * @return The split times for the report.
     */
    public Object[][] getSplitTimes() {
        calculateSplitTimes();
        return splitTimes;
    }

    /**
     * Method to generate a report for a specific channel.
     * @param chan The channel you want to use.
     * @param type The type of report (1 = avg, 2 = max, 3 = min)
     * @return The array of objects to use in the table.
     */
    public Object[][] getSegmentValues(Channel chan, int type) {
        double[] tempValues = chan.getCalibratedValues();
        final int amountOfLaps = segmentation.getAmountOfLaps();
        final int amountOfSegments = segmentation.getAmountOfSegments();
        Object[][] segmentValues = new Object[amountOfLaps][amountOfSegments + 2];
        double temp;
        double currentValue;
        int currentLap = 1;
        int currentSegment = 1;
        int lastOffset = 0;
        int nextOffset = segmentation.getSegmentOffset(currentLap, currentSegment);
        int i = 0;
        if (type == 1) {
            temp = 0.0;
        } else if (type == 2) {
            temp = Double.MIN_VALUE;
        } else {
            temp = Double.MAX_VALUE;
        }


        while (i < tempValues.length && currentLap <= amountOfLaps) {
            currentValue = tempValues[i];
            if (type == 1) {
                temp += tempValues[i];
            } else if (type == 2 && currentValue > temp) {
                temp = currentValue;
            } else if (type == 3 && currentValue < temp) {
                temp = currentValue;
            }

            if (i == nextOffset) {
                if (currentSegment == 1) {
                    segmentValues[currentLap - 1][0] = "Lap # " + currentLap;
                }
                if (type == 1) {
                    segmentValues[currentLap - 1][currentSegment] = MathFunction.roundTwoDecimals(temp / (nextOffset - lastOffset));
                    temp = 0.0;
                } else if (type == 2 || type == 3) {
                    segmentValues[currentLap - 1][currentSegment] = MathFunction.roundTwoDecimals(temp);
                    if (type == 2) {
                        temp = Double.MIN_VALUE;
                    } else {
                        temp = Double.MAX_VALUE;
                    }
                }

                if (currentSegment == amountOfSegments + 1) {
                    currentLap++;
                    currentSegment = 1;
                } else {
                    currentSegment++;
                }
                lastOffset = nextOffset + 1;
                nextOffset = segmentation.getSegmentOffset(currentLap, currentSegment);
            }
            i++;
        }
        return segmentValues;
    }
}
