/**
 * Class that implements the track map plugin prototype.
 * @author Lauritz Møllersen
 */
package plugin.report;

import channel.Channel;
import error.ErrorLogger;
import java.awt.Component;
import java.awt.GridLayout;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import plugin.DefaultPlugin;
import plugin.Plugin;
import parser.Constants;
import project.Project;
import segmentation.Segmentation;
import view.View;
import error.Error;
import plugin.PluginInitException;

/**
 *
 * @author Per Øyvind
 */
public class ReportPlugin extends DefaultPlugin {

    private static final long serialVersionUID = 1L;
    private Project p;
    private Segmentation s;
    private ReportPropPanel propPanel;
    private Report report;
    private JPanel jpanel;
    private ReportTable reportTable;
    private JScrollPane scrollPane;
    private int[] channels; // not in use

    /**
     * 
     */
    public ReportPlugin() {
    }

    /**
     * Initializes the plugin by setting view reference and plugin ID. This override
     * also contains specific initialization code for this plugin.
     * @param v The parent view of the plugin
     * @throws PluginInitException  
     */
    @Override
    public void init(View v) throws PluginInitException {
        super.init(v);
        p = v.getParentProject();
        s = p.getSegmentation();
        Channel tempLap = p.getChannelType(Constants.CHAN_LAP);
        if (tempLap == null) {
            ErrorLogger.addError(new Error("Channel error", "Error with the lap channel. "
                    + "Check that you have defined this in the Channel Options."));
            throw new PluginInitException();
        } else if (s == null || s.getAmountOfLaps() < 1 || s.getAmountOfSegments() < 1) {
            ErrorLogger.addError(new Error("Error with segmentation", "Check that you have defined the segments with the Map Plugin."));
            throw new PluginInitException();
        }



        report = new Report(s, p.getChannelType(Constants.CHAN_TIME));

        propPanel = new ReportPropPanel(this);

        jpanel = new JPanel(new GridLayout(1, 0));


        //Create and set up the content pane.
        reportTable = new ReportTable(report.getHeaders(), report.getSplitTimes());

        //Create the scroll pane and add the table to it.
        scrollPane = new JScrollPane(reportTable);

        //Add the scroll pane to this panel.
        jpanel.add(scrollPane);


    }

    /**
     * 
     * @param channelName
     * @param type
     */
    public void showChannelReport(String channelName, int type) {
        Channel tempChan = p.getChannel(channelName);
        if (tempChan != null) {
            reportTable.reCalc(report.getSegmentValues(tempChan, type), false);
        }
    }

    /**
     * 
     */
    public void showSplitTimes() {
        reportTable.reCalc(report.getSplitTimes(), true);
    }

    /**
     * 
     * @return
     */
    public String[] getChanStrings() {
        return p.getChanStrings();
    }

    /**
     * Get the main GUI component of the Plugin.
     * @return GUI component.
     */
    @Override
    public JPanel getPluginPanel() {
        return jpanel;
    }

    /**
     * Get the property GUI component of the Plugin. 
     * @return GU /**I component.
     */
    public Component getPropertiesPanel() {
        return this.propPanel;
    }

    /**
     * Sets a property in the plugin with undo functionality.
     * This override also takes care of updating the GUI components when a 
     * property is changed.
     * @return  
     */
    public int[] getChannels() {
        return this.channels;
    }

    /**
     * 
     * @param channels
     * @param undo
     */
    public void setChannels(int[] channels, boolean undo) {
        this.channels = channels;
    }

    /**
     * 
     * @param title
     * @param undo
     */
    public void setTitle(String title, boolean undo) {
        super.setTitle(title, undo);
        //Update GUI
        if (propPanel != null) {
            propPanel.updateTitle();
        }
    }

    /**
     * 
     * @return
     */
    @Override
    public boolean isCursorLinked() {
        return false;
    }
}
