package plugin.timeslip;

import channel.Channel;
import error.ErrorLogger;
import java.awt.Component;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.ArrayList;
import javax.swing.JPanel;
import javax.swing.ProgressMonitor;
import plugin.DefaultPlugin;
import plugin.PluginInitException;
import project.Command;
import view.View;
import parser.Constants;
import segmentation.Segmentation;

/**
 * Class that implements the timeslip plugin prototype.
 * @author Per Øyvind Stadheim
 */
public class TimeSlipPlugin extends DefaultPlugin {

    private static final long serialVersionUID = 1L;
    private int dataVersion = 1;
    private int baseLap = -1;
    private ArrayList<Integer> selectedLaps;
    private ArrayList<Integer> selectedSegments;
    private boolean followDomainCursor = true;
    private PluginPanel pluginPanel;
    private TabbedPropPanel propPanel;
    private Channel timeDifferenceChannel;

    /**
     * 
     */
    public TimeSlipPlugin() {
    }

    /**
     * Initializes the plugin by setting view reference and plugin ID. This override
     * also contains specific initialization code for this plugin.
     * @param v The parent view of the plugin
     * @throws PluginInitException  
     */
    @Override
    public void init(View v) throws PluginInitException {
        super.init(v);
        //Check if dataset supports plugin
        if (getParentProject().getChannelType(Constants.CHAN_DISTANCE) == null
                || getParentProject().getChannelType(Constants.CHAN_TIME) == null) {
            ErrorLogger.addError(new error.Error("Channel error", "Error with the time and distance channels. "
                    + "Check that you have defined these in the Channel Options."));
            throw new PluginInitException();
        }


        //If not loaded from file, intialize default fields.

        if (selectedLaps == null) {
            selectedLaps = new ArrayList<Integer>();
        }
        if (selectedSegments == null) {
            selectedSegments = new ArrayList<Integer>();
        }

        //Check and eventually nerf plugin settings vs current project (segmentation, channels)
        //  removeUnsupportedChannelValues();
        //removeUnsupportedLapValues();
        //removeUnsupportedSegmentValues();

        //Create GUI components
        propPanel = new TabbedPropPanel();
        propPanel.setTsp(this);
        pluginPanel = new PluginPanel(this);
        propPanel.setup();
        pluginPanel.generateDomainSegmentOffsets();
        if (!selectedLaps.isEmpty()) {
            pluginPanel.rebuildSeries();
        }
    }

    // Methods used internally
    private void removeUnsupportedLapValues() {
        for (int i = 0; i < selectedLaps.size(); i++) {
            boolean exists = false;
            if (selectedLaps.get(i) <= getParentProject().getSegmentation().getAmountOfLaps()) {
                exists = true;
            }
            if (!exists) {
                selectedLaps.remove(i);
                i--;
            }
        }
    }

    private void removeUnsupportedSegmentValues() {
        for (int i = 0; i < selectedSegments.size(); i++) {
            boolean exists = false;
            if (selectedSegments.get(i) <= getParentProject().getSegmentation().getAmountOfSegments() + 1) {
                exists = true;
            }
            if (!exists) {
                selectedSegments.remove(i);
                i--;
            }
        }
    }

    /**
     * 
     * @param b
     * @param undo
     */
    @Override
    public void setCursorLinked(boolean b, boolean undo) {
        super.setCursorLinked(b, undo);
        propPanel.linkCursorChanged();
    }

    /**
     * 
     * @param lap
     * @param value
     * @param undo
     */
    public void setLapSelection(int lap, boolean value, boolean undo) {
        if (!undo) {
            if (!value && selectedLaps.contains(new Integer(lap))) {
                selectedLaps.remove(new Integer(lap));
            } else if (value && !selectedLaps.contains(lap)) {
                selectedLaps.add(lap);
            }

            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.lapsChanged();
        } else {
            final int targetLap = lap;
            final boolean oldValue = selectedLaps.contains(lap);
            final boolean newValue = value;
            final TimeSlipPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setLapSelection(targetLap, newValue, false);
                }

                @Override
                public void reverse() {
                    source.setLapSelection(targetLap, oldValue, false);
                }

                @Override
                public String toString() {
                    return "Lap Selection";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public ArrayList<Integer> getSelectedLaps() {
        return this.selectedLaps;
    }

    /**
     * 
     * @param segment
     * @param value
     * @param undo
     */
    public void setSegmentSelection(int segment, boolean value, boolean undo) {
        if (!undo) {
            if (!value && selectedSegments.contains(new Integer(segment))) {
                selectedSegments.remove(new Integer(segment));
            } else if (value && !selectedSegments.contains(segment)) {
                selectedSegments.add(segment);
            }
            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.segmentsChanged();
        } else {
            final int targetSegment = segment;
            final boolean oldValue = selectedSegments.contains(segment);
            final boolean newValue = value;
            final TimeSlipPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setSegmentSelection(targetSegment, newValue, false);
                }

                @Override
                public void reverse() {
                    source.setSegmentSelection(targetSegment, oldValue, false);
                }

                @Override
                public String toString() {
                    return "Segment Selection";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public ArrayList<Integer> getSelectedSegments() {
        return this.selectedSegments;
    }

    /**
     * 
     * @param b
     * @param undo
     */
    public void setFollowDomainCursor(boolean b, boolean undo) {
        if (!undo) {
            this.followDomainCursor = b;
            //Update GUI
            pluginPanel.setFollowDomainCursor(b);
            propPanel.followCursorChanged();
        } else {
            final boolean oldValue = this.isFollowDomainCursor();
            final boolean newValue = b;
            final TimeSlipPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setFollowDomainCursor(newValue, false);
                }

                @Override
                public void reverse() {
                    source.setFollowDomainCursor(oldValue, false);
                }

                @Override
                public String toString() {
                    return "Set Follow Cursor";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public boolean isFollowDomainCursor() {
        return followDomainCursor;
    }

    /**
     * 
     * @param title
     * @param undo
     */
    @Override
    public void setTitle(String title, boolean undo) {
        super.setTitle(title, undo);

        //Update GUI
        if (propPanel != null) {
            propPanel.titleChanged();
        }

    }

    private int findClosestLapRelativeIndex(double distance, int lap, int memoindex) {
        Segmentation s = getParentProject().getSegmentation();
        int fromIndex = s.getLapOffset(lap);
        int toIndex = s.getLapEndOffset(lap) + 1;
        double lapOffsetDistance = getParentProject().getChannelType(Constants.CHAN_DISTANCE).getCalibratedValues()[fromIndex];
        //  int searches = 0;

        double closestDistance = Double.POSITIVE_INFINITY;
        int closestIndex = -1;
        //i=fromIndex
        for (int i = memoindex; i < toIndex; i++) {
            double tempdist = Math.abs((distance + lapOffsetDistance) - getParentProject().getChannelType(Constants.CHAN_DISTANCE).getCalibratedValues()[i]);
            if (Math.abs(tempdist) <= closestDistance) {
                closestDistance = Math.abs(tempdist);
                closestIndex = i;


            }
            // searches++;
            //  if(searches)return closestIndex;
        }
        return closestIndex;
        //  return quickfind(lapOffsetDistance, fromIndex, toIndex, getParentProject().getChannelType(Constants.CHAN_DISTANCE).getCalibratedValues());

    }

    /**
     * 
     * @param baseLap
     */
    public void generateTimeDifferenceChannel(int baseLap) {

        Segmentation s = getParentProject().getSegmentation();
        Channel time = getParentProject().getChannelType(Constants.CHAN_TIME);
        Channel distance = getParentProject().getChannelType(Constants.CHAN_DISTANCE);
        double[] tempValues = new double[time.getCalibratedValues().length];
        ProgressMonitor pbar = new ProgressMonitor(null, "Calculating time differences..",
                "Progress", 0, time.getCalibratedValues().length);
        pbar.setMillisToDecideToPopup(1);
        pbar.setProgress(0);
        for (int lap = 1; lap < getParentProject().getSegmentation().getAmountOfLaps() + 1; lap++) {

            int memoindex = s.getLapOffset(baseLap);;
            int lapOffset = s.getLapOffset(lap);

            int lapEndOffset = s.getLapEndOffset(lap) + 1;

            for (int j = lapOffset; j < lapEndOffset; j++) {
                pbar.setProgress(j);
                double dist = distance.getCalibratedValues()[j] - distance.getCalibratedValues()[lapOffset];

                int baseLapIndex = findClosestLapRelativeIndex(dist, baseLap, memoindex);
                memoindex = baseLapIndex;

                if (baseLapIndex >= 0) {
                    tempValues[j] = (time.getCalibratedValues()[baseLapIndex] - time.getCalibratedValues()[s.getLapOffset(baseLap)])
                            - (time.getCalibratedValues()[j] - time.getCalibratedValues()[lapOffset]);
                }
            }

        }
        pbar.setProgress(time.getCalibratedValues().length);
        timeDifferenceChannel = new Channel(tempValues, "timeslip");

    }
    // Methods used externally

    /**
     * 
     */
    @Override
    public void colorsChanged() {
        pluginPanel.rebuildSeries();
        propPanel.lapColorsChanged();
    }

    /**
     * 
     */
    @Override
    public void segmentationChanged() {
        //generateTimeDifferenceChannel(baseLap);
        pluginPanel.rebuildSeries();
        pluginPanel.generateDomainSegmentOffsets();
        propPanel.lapsChanged();
        propPanel.segmentsChanged();
    }

    /**
     * 
     * @param position
     */
    @Override
    public synchronized void setCursorPosition(int position) {
        super.setCursorPosition(position);
        //Updating GUI
        pluginPanel.setCursorPosition(position);
    }

    /**
     * Get the main GUI component of the Plugin.
     * @return GUI component.
     */
    @Override
    public JPanel getPluginPanel() {
        return this.pluginPanel;

    }

    /**
     * Get the property GUI component of the Plugin. 
     * @return GU /**I component.
     */
    @Override
    public Component getPropertiesPanel() {
        return this.propPanel;
    }

    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        removeUnsupportedLapValues();
        removeUnsupportedSegmentValues();
        super.writeExternal(out);
        out.writeInt(this.dataVersion);
        out.writeInt(this.baseLap);
        out.writeObject(this.getSelectedLaps());
        out.writeObject(this.getSelectedSegments());
        out.writeBoolean(followDomainCursor);

    }

    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        super.readExternal(in);
        int version = in.readInt();
        if (version == 1) {
            this.baseLap = in.readInt();
            this.selectedLaps = (ArrayList<Integer>) in.readObject();
            this.selectedSegments = (ArrayList<Integer>) in.readObject();
            this.followDomainCursor = in.readBoolean();

        }
    }

    /**
     * 
     * @return
     */
    public int getBaseLap() {
        return this.baseLap;
    }

    /**
     * 
     * @param lap
     * @param undo
     */
    public void setBaseLap(int lap, boolean undo) {
        if (!undo) {
            this.baseLap = lap;
            generateTimeDifferenceChannel(lap);
            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.lapsChanged();

        } else {
            final int newValue = lap;
            final int oldValue = this.baseLap;
            final TimeSlipPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setBaseLap(newValue, false);
                }

                @Override
                public void reverse() {
                    source.setBaseLap(oldValue, false);
                }

                @Override
                public String toString() {
                    return "Set Base Lap";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public Channel getTimeDifferenceChannel() {
        return this.timeDifferenceChannel;
    }
}
