/*
 * This file is part of VLCJ.
 *
 * VLCJ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VLCJ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VLCJ.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright 2009, 2010, 2011 Caprica Software Limited.
 */
package plugin.video;

import java.awt.AWTEvent;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.AWTEventListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.JPopupMenu;

import uk.co.caprica.vlcj.binding.internal.libvlc_media_t;
import uk.co.caprica.vlcj.logger.Logger;
import uk.co.caprica.vlcj.player.AudioOutput;
import uk.co.caprica.vlcj.player.MediaDetails;
import uk.co.caprica.vlcj.player.MediaPlayer;
import uk.co.caprica.vlcj.player.MediaPlayerEventAdapter;
import uk.co.caprica.vlcj.player.MediaPlayerFactory;
import uk.co.caprica.vlcj.player.embedded.EmbeddedMediaPlayer;
//import uk.co.caprica.vlcj.player.events.VideoOutputEventListener;
import uk.co.caprica.vlcj.runtime.RuntimeUtil;
import uk.co.caprica.vlcj.runtime.windows.WindowsCanvas;

import com.sun.awt.AWTUtilities;
import com.sun.jna.NativeLibrary;
import com.sun.jna.platform.WindowUtils;
import error.ErrorLogger;
import java.awt.BorderLayout;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.util.Properties;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;
import plugin.DefaultPlugin;
import plugin.Plugin;
import plugin.PluginInitException;
import parser.Constants;
import project.Project;
import uk.co.caprica.vlcj.binding.LibVlc;
import uk.co.caprica.vlcj.binding.LibVlcFactory;
import uk.co.caprica.vlcj.filter.swing.SwingFileFilterFactory;
import uk.co.caprica.vlcj.runtime.x.LibXUtil;
import view.View;

/**
 * VideoPlugin for the Revolver Analyzer project.
 */
public class VideoPlugin extends DefaultPlugin {

    private static final long serialVersionUID = 1L;
    private long timeOffset = 0;
    private String vlcDir;
    private VideoPropPanel vp;
    private JFileChooser fileChooser;
    private Project p;
    private double[] tc;
    private long currentTime = 0;
    private int counter = 0;
    private JPanel mainFrame;
    private Canvas videoSurface;
    private JPanel controlsPanel;
    private JPanel videoAdjustPanel;
    private MediaPlayerFactory mediaPlayerFactory;
    private EmbeddedMediaPlayer mediaPlayer;
    private static final String VLCJ_LOG_LEVEL = "INFO";
    /**
     * Change this to point to your own vlc installation, or comment out the code
     * if you want to use your system default installation.
     * <p>
     * This is a bit more explicit than using the -Djna.library.path= system 
     * property.
     */
    private static final String NATIVE_LIBRARY_SEARCH_PATH = null;
    /**
     * Set to true to dump out native JNA memory structures.
     */
    private static final String DUMP_NATIVE_MEMORY = "false";

    /**
     * Static initialisation.
     */
    static {
        if (null == System.getProperty("vlcj.log")) {
            System.setProperty("vlcj.log", VLCJ_LOG_LEVEL);
        }

        // Safely try to initialise LibX11 to reduce the opportunity for native
        // crashes - this will silently throw an Error on Windows (and maybe MacOS)
        // that can safely be ignored
        LibXUtil.initialise();

        if (null != NATIVE_LIBRARY_SEARCH_PATH) {
            Logger.info("Explicitly adding JNA native library search path: '{}'", NATIVE_LIBRARY_SEARCH_PATH);
            NativeLibrary.addSearchPath(RuntimeUtil.getLibVlcLibraryName(), NATIVE_LIBRARY_SEARCH_PATH);
        }

        System.setProperty("jna.dump_memory", DUMP_NATIVE_MEMORY);
    }

    /**
     * 
     */
    public VideoPlugin() {
    }

    @Override
    public void init(View v) throws PluginInitException {
        super.init(v);
        p = v.getParentProject();
        if (p.getChannelType(parser.Constants.CHAN_TIME) == null) {
            ErrorLogger.addError(new error.Error("Data error", "No data has been imported yet."));
            throw new PluginInitException();
        }
        tc = p.getChannelType(Constants.CHAN_TIME).getCalibratedValues();

        fileChooser = new JFileChooser();
        fileChooser.setApproveButtonText("Play");
        fileChooser.addChoosableFileFilter(SwingFileFilterFactory.newVideoFileFilter());
        fileChooser.addChoosableFileFilter(SwingFileFilterFactory.newAudioFileFilter());
        fileChooser.addChoosableFileFilter(SwingFileFilterFactory.newPlayListFileFilter());
        FileFilter defaultFilter = SwingFileFilterFactory.newMediaFileFilter();
        fileChooser.addChoosableFileFilter(defaultFilter);
        fileChooser.setFileFilter(defaultFilter);

        // read videoplugin options
        File file = new File("video.properties");
        Properties videoProps = new Properties();
        vlcDir = "";
        try {
            FileInputStream fis = new FileInputStream(file);
            videoProps.load(fis);
            vlcDir = videoProps.getProperty("vlcdir");
            if (vlcDir == null) {
                throw new Exception();
            }
        } catch (Exception e) {

            try {
                FileWriter fw = new FileWriter("video.properties");
                JFileChooser fc = new JFileChooser("c:/Program Files/");
                fc.setDialogTitle("Select your VLC directory");
                fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
                int returnVal = fc.showOpenDialog(mainFrame);

                if (returnVal == JFileChooser.APPROVE_OPTION) {
                    File vlcFile = fc.getSelectedFile();
                    vlcDir = vlcFile.getAbsolutePath();
                    videoProps.setProperty("vlcdir", vlcDir);
                    videoProps.store(fw, "Properties for the Video Plugin in RevolveAnalyze");
                }
            } catch (Exception ex) {
                error.ErrorLogger.addError(new error.Error(ex, "Video plugin options", "Problem saving the video options to a file."));
            }
        }


        NativeLibrary.addSearchPath("libvlc", vlcDir);
        System.setProperty("jna.library.path", vlcDir);
        LibVlc libVlc = LibVlcFactory.factory().create();

        Logger.info("  version: {}", libVlc.libvlc_get_version());
        Logger.info(" compiler: {}", libVlc.libvlc_get_compiler());
        Logger.info("changeset: {}", libVlc.libvlc_get_changeset());

        if (RuntimeUtil.isWindows()) {
            // If running on Windows and you want the mouse/keyboard event hack...
            videoSurface = new WindowsCanvas();
        } else {
            videoSurface = new Canvas();
        }

        Logger.debug("videoSurface={}", videoSurface);

        videoSurface.setBackground(Color.black);
        //videoSurface.setSize(400, 300); // Only for initial layout

        // Since we're mixing lightweight Swing components and heavyweight AWT 
        // components this is probably a good idea
        JPopupMenu.setDefaultLightWeightPopupEnabled(false);

        TestPlayerMouseListener mouseListener = new TestPlayerMouseListener();
        videoSurface.addMouseListener(mouseListener);
        videoSurface.addMouseMotionListener(mouseListener);
        videoSurface.addMouseWheelListener(mouseListener);
        videoSurface.addKeyListener(new TestPlayerKeyListener());

        List<String> vlcArgs = new ArrayList<String>();

//    vlcArgs.add("--ffmpeg-hw"); // <--- if your system supports it, this might be useful
        vlcArgs.add("--no-plugins-cache");
        vlcArgs.add("--no-video-title-show");
        vlcArgs.add("--no-snapshot-preview");
        vlcArgs.add("--quiet");
        vlcArgs.add("--quiet-synchro");
        vlcArgs.add("--intf");
        vlcArgs.add("dummy");

        // Special case to help out users on Windows (supposedly this is not actually needed)...
//    if(RuntimeUtil.isWindows()) {
//      vlcArgs.add("--plugin-path=" + WindowsRuntimeUtil.getVlcInstallDir() + "\\plugins");
//    }
//    else {
//      vlcArgs.add("--plugin-path=/home/linux/vlc/lib");
//    }

//    vlcArgs.add("--plugin-path=" + System.getProperty("user.home") + "/.vlcj");

        Logger.debug("vlcArgs={}", vlcArgs);

        mainFrame = new JPanel();

        mediaPlayerFactory = new MediaPlayerFactory(vlcArgs.toArray(new String[vlcArgs.size()]));
        mediaPlayerFactory.setUserAgent("vlcj test player");

        List<AudioOutput> audioOutputs = mediaPlayerFactory.getAudioOutputs();
        Logger.debug("audioOutputs={}", audioOutputs);

        mediaPlayer = mediaPlayerFactory.newEmbeddedMediaPlayer();
        getMediaPlayer().setVideoSurface(mediaPlayerFactory.newVideoSurface(videoSurface));
        getMediaPlayer().setPlaySubItems(true);

        getMediaPlayer().setEnableKeyInputHandling(false);
        getMediaPlayer().setEnableMouseInputHandling(false);

        controlsPanel = new PlayerControlsPanel(getMediaPlayer());
        videoAdjustPanel = new PlayerVideoAdjustPanel(getMediaPlayer());

        mainFrame.setPreferredSize(new Dimension(800, 600));
        mainFrame.setMinimumSize(new Dimension(1, 1));
        mainFrame.setLayout(new BorderLayout());
        mainFrame.setBackground(Color.black);
        mainFrame.add(videoSurface, BorderLayout.CENTER);
        mainFrame.add(controlsPanel, BorderLayout.SOUTH);

        // Global AWT key handler, you're better off using Swing's InputMap and 
        // ActionMap with a JFrame - that would solve all sorts of focus issues too
        Toolkit.getDefaultToolkit().addAWTEventListener(new AWTEventListener() {

            @Override
            public void eventDispatched(AWTEvent event) {
                if (event instanceof KeyEvent) {
                    KeyEvent keyEvent = (KeyEvent) event;
                    if (keyEvent.getID() == KeyEvent.KEY_PRESSED) {
                        if (keyEvent.getKeyCode() == KeyEvent.VK_F12) {
                            controlsPanel.setVisible(!controlsPanel.isVisible());
                            videoAdjustPanel.setVisible(!videoAdjustPanel.isVisible());
                            mainFrame.invalidate();
                            mainFrame.validate();
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_A) {
                            getMediaPlayer().setAudioDelay(getMediaPlayer().getAudioDelay() - 50000);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_S) {
                            getMediaPlayer().setAudioDelay(getMediaPlayer().getAudioDelay() + 50000);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_1) {
                            getMediaPlayer().setTime(60000 * 1);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_2) {
                            getMediaPlayer().setTime(60000 * 2);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_3) {
                            getMediaPlayer().setTime(60000 * 3);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_4) {
                            getMediaPlayer().setTime(60000 * 4);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_5) {
                            getMediaPlayer().setTime(60000 * 5);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_6) {
                            getMediaPlayer().setTime(60000 * 6);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_7) {
                            getMediaPlayer().setTime(60000 * 7);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_8) {
                            getMediaPlayer().setTime(60000 * 8);
                        } else if (keyEvent.getKeyCode() == KeyEvent.VK_9) {
                            getMediaPlayer().setTime(60000 * 9);
                        }
                    }
                }
            }
        }, AWTEvent.KEY_EVENT_MASK);

        mainFrame.setVisible(true);

        getMediaPlayer().addMediaPlayerEventListener(new TestPlayerMediaPlayerEventListener());

        // Won't work with OpenJDK or JDK1.7, requires a Sun/Oracle JVM (currently)
        boolean transparentWindowsSupport = true;
        try {
            Class.forName("com.sun.awt.AWTUtilities");
        } catch (Exception e) {
            transparentWindowsSupport = false;
        }

        Logger.debug("transparentWindowsSupport={}", transparentWindowsSupport);

        if (transparentWindowsSupport) {
            final Window test = new Window(null, WindowUtils.getAlphaCompatibleGraphicsConfiguration()) {

                private static final long serialVersionUID = 1L;

                public void paint(Graphics g) {
                    Graphics2D g2 = (Graphics2D) g;

                    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
                    g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_LCD_HRGB);

                    g.setColor(Color.white);
                    g.fillRoundRect(100, 150, 100, 100, 32, 32);

                    g.setFont(new Font("Sans", Font.BOLD, 32));
                    g.drawString("Heavyweight overlay test", 100, 300);
                }
            };

            AWTUtilities.setWindowOpaque(test, false); // Doesn't work in full-screen exclusive mode, you would have to use 'simulated' full-screen - requires Sun/Oracle JDK
            test.setBackground(new Color(0, 0, 0, 0)); // This is what you do in JDK7
        }

        // This might be useful
//	enableMousePointer(false);

        vp = new VideoPropPanel(this);


    }

    @Override
    public JPanel getPropertiesPanel() {
        return vp;
    }

    @Override
    public JPanel getPluginPanel() {
        return mainFrame;
    }

    /**
     * 
     */
    public void playMovie() {
        getMediaPlayer().playMedia("D:/film.avi");
        getMediaPlayer().enableOverlay(true);
    }

    /**
     * 
     */
    public void openMovie() {
        getMediaPlayer().enableOverlay(false);
        if (JFileChooser.APPROVE_OPTION == fileChooser.showOpenDialog(controlsPanel)) {
            getMediaPlayer().playMedia(fileChooser.getSelectedFile().getAbsolutePath());
        }
        getMediaPlayer().enableOverlay(true);
    }

    /**
     * 
     * @param time
     */
    public void setTime(double time) {
        getMediaPlayer().setTime(timeOffset + (long) (time * 1000));
    }

    /**
     * @return the mediaPlayer
     */
    public EmbeddedMediaPlayer getMediaPlayer() {
        return mediaPlayer;
    }

    /**
     * @param mediaPlayer the mediaPlayer to set
     */
    public void setMediaPlayer(EmbeddedMediaPlayer mediaPlayer) {
        this.mediaPlayer = mediaPlayer;
    }

    /**
     * Log level, used only if the -Dvlcj.log= system property has not already 
     * been set.
     */
    private final class TestPlayerMediaPlayerEventListener extends MediaPlayerEventAdapter {

        @Override
        public void mediaChanged(MediaPlayer mediaPlayer) {
            Logger.debug("mediaChanged(mediaPlayer={})", mediaPlayer);
        }

        @Override
        public void finished(MediaPlayer mediaPlayer) {
            Logger.debug("finished(mediaPlayer={})", mediaPlayer);
        }

        @Override
        public void paused(MediaPlayer mediaPlayer) {
            Logger.debug("paused(mediaPlayer={})", mediaPlayer);
        }

        @Override
        public void playing(MediaPlayer mediaPlayer) {
            Logger.debug("playing(mediaPlayer={})", mediaPlayer);
            MediaDetails mediaDetails = mediaPlayer.getMediaDetails();
            Logger.info("mediaDetails={}", mediaDetails);
        }

        @Override
        public void stopped(MediaPlayer mediaPlayer) {
            Logger.debug("stopped(mediaPlayer={})", mediaPlayer);
        }

        @Override
        public void videoOutput(MediaPlayer mediaPlayer, int newCount) {
            Logger.debug("videoOutput(mediaPlayer={},newCount={})", mediaPlayer, newCount);
        }

        @Override
        public void error(MediaPlayer mediaPlayer) {
            Logger.debug("error(mediaPlayer={})", mediaPlayer);
        }

        @Override
        public void mediaSubItemAdded(MediaPlayer mediaPlayer, libvlc_media_t subItem) {
            Logger.debug("mediaSubItemAdded(mediaPlayer={},subItem={})", mediaPlayer, subItem);
        }

        @Override
        public void mediaDurationChanged(MediaPlayer mediaPlayer, long newDuration) {
            Logger.debug("mediaDurationChanged(mediaPlayer={},newDuration={})", mediaPlayer, newDuration);
        }

        @Override
        public void mediaParsedChanged(MediaPlayer mediaPlayer, int newStatus) {
            Logger.debug("mediaParsedChanged(mediaPlayer={},newStatus={})", mediaPlayer, newStatus);
        }

        @Override
        public void mediaFreed(MediaPlayer mediaPlayer) {
            Logger.debug("mediaFreed(mediaPlayer={})", mediaPlayer);
        }

        @Override
        public void mediaStateChanged(MediaPlayer mediaPlayer, int newState) {
            Logger.debug("mediaStateChanged(mediaPlayer={},newState={})", mediaPlayer, newState);
        }

        @Override
        public void mediaMetaChanged(MediaPlayer mediaPlayer, int metaType) {
            Logger.debug("mediaMetaChanged(mediaPlayer={},metaType={})", mediaPlayer, metaType);
        }
    }

    /**
     * 
     * 
     * @param enable
     */
    @SuppressWarnings("unused")
    private void enableMousePointer(boolean enable) {
        Logger.debug("enableMousePointer(enable={})", enable);
        if (enable) {
            videoSurface.setCursor(null);
        } else {
            Image blankImage = new BufferedImage(1, 1, BufferedImage.TYPE_INT_ARGB);
            videoSurface.setCursor(Toolkit.getDefaultToolkit().createCustomCursor(blankImage, new Point(0, 0), ""));
        }
    }

    /**
     *
     */
    private final class TestPlayerMouseListener extends MouseAdapter {

        @Override
        public void mouseMoved(MouseEvent e) {
            Logger.trace("mouseMoved(e={})", e);
        }

        @Override
        public void mousePressed(MouseEvent e) {
            Logger.debug("mousePressed(e={})", e);
        }

        @Override
        public void mouseReleased(MouseEvent e) {
            Logger.debug("mouseReleased(e={})", e);
        }

        @Override
        public void mouseWheelMoved(MouseWheelEvent e) {
            Logger.debug("mouseWheelMoved(e={})", e);
        }

        @Override
        public void mouseEntered(MouseEvent e) {
            Logger.debug("mouseEntered(e={})", e);
        }

        @Override
        public void mouseExited(MouseEvent e) {
            Logger.debug("mouseExited(e={})", e);
        }
    }

    /**
     *
     */
    private final class TestPlayerKeyListener extends KeyAdapter {

        @Override
        public void keyPressed(KeyEvent e) {
            Logger.debug("keyPressed(e={})", e);
        }

        @Override
        public void keyReleased(KeyEvent e) {
            Logger.debug("keyReleased(e={})", e);
        }

        @Override
        public void keyTyped(KeyEvent e) {
            Logger.debug("keyTyped(e={})", e);
        }
    }

    /**
     * 
     * @param title
     * @param undo
     */
    public void setTitle(String title, boolean undo) {
        super.setTitle(title, undo);
        /* if (propPanel != null) {
        propPanel.updateTitle();*/
    }

    /**
     * 
     * @param offset
     */
    public void setTimeoffset(long offset) {
        timeOffset = offset;
    }

    /**
     * 
     * @param pos
     */
    @Override
    public void setCursorPosition(int pos) {
        if (counter % 10 == 0) {
            final long globalTime = (long) (tc[pos] * 1000) + timeOffset;
            final long tempTime = getMediaPlayer().getTime();
            if (globalTime > 0 && Math.abs(globalTime - tempTime) > 2000) {
                getMediaPlayer().setTime(globalTime);
            }
        }
        counter++;
    }

    /**
     * 
     */
    @Override
    public void startPlayback() {
        getMediaPlayer().play();
    }

    /**
     * 
     */
    @Override
    public void stopPlayback() {
        getMediaPlayer().stop();
    }
}
