package plugin.xy;

import error.ErrorLogger;
import plugin.Plugin;
import java.awt.Component;
import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.ArrayList;
import javax.swing.JPanel;
import plugin.DefaultPlugin;
import plugin.PluginInitException;
import project.Command;
import view.View;
import parser.Constants;

/**
 * Class that implements the xychart plugin prototype.
 * @author Per Øyvind Stadheim
 */
public class XYChartPlugin extends DefaultPlugin {

    private static final long serialVersionUID = 1L;
    private int dataVersion = 1;
    private String xAxis;
    private String yAxis;
    private ArrayList<Integer> selectedLaps;
    private ArrayList<Integer> selectedSegments;
    private boolean followDomainCursor = true;
    private PluginPanel pluginPanel;
    private PropertiesPanel propPanel;

    /**
     * 
     */
    public XYChartPlugin() {
    }

    /**
     * Initializes the plugin by setting view reference and plugin ID. This override
     * also contains specific initialization code for this plugin.
     * @param v The parent view of the plugin
     * @throws PluginInitException  
     */
    @Override
    public void init(View v) throws PluginInitException {
        super.init(v);
        //Check if dataset supports plugin
        if (getParentProject().getChannelType(Constants.CHAN_DISTANCE) == null
                || getParentProject().getChannelType(Constants.CHAN_TIME) == null) {
            ErrorLogger.addError(new error.Error("Channel error", "Error with the time and distance channels. "
                    + "Check that you have defined these in the Channel Options."));
            throw new PluginInitException();
        }


        //If not loaded from file, intialize default fields.

        if (selectedLaps == null) {
            selectedLaps = new ArrayList<Integer>();
        }
        if (selectedSegments == null) {
            selectedSegments = new ArrayList<Integer>();
        }

        //Check and eventually nerf plugin settings vs current project (segmentation, channels)
        removeUnsupportedChannelValues();
        removeUnsupportedLapValues();
        removeUnsupportedSegmentValues();

        //Create GUI components
        propPanel = new PropertiesPanel(this);
        pluginPanel = new PluginPanel(this);
        propPanel.setup();
        //pluginPanel.generateDomainSegmentOffsets();
        if (xAxis != null) {
            pluginPanel.rebuildSeries();
        }
    }

    // Methods used internally
    /**
     * 
     */
    public void removeUnsupportedLapValues() {
        for (int i = 0; i < selectedLaps.size(); i++) {
            boolean exists = false;
            if (selectedLaps.get(i) <= getParentProject().getSegmentation().getAmountOfLaps()) {
                exists = true;
            }
            if (!exists) {
                selectedLaps.remove(i);
                i--;
            }
        }
    }

    /**
     * 
     */
    public void removeUnsupportedSegmentValues() {
        for (int i = 0; i < selectedSegments.size(); i++) {
            boolean exists = false;
            if (selectedSegments.get(i) <= getParentProject().getSegmentation().getAmountOfSegments() + 1) {
                exists = true;
            }
            if (!exists) {
                selectedSegments.remove(i);
                i--;
            }
        }
    }

    /**
     * 
     */
    public void removeUnsupportedChannelValues() {
        boolean xExists = false;
        boolean yExists = false;
        for (String s2 : getParentProject().getChanStrings()) {
            if (xAxis != null && xAxis.equals(s2)) {
                xExists = true;
            }
            if (yAxis != null && yAxis.equals(s2)) {
                yExists = true;
            }
        }
        for (String s3 : getParentProject().getMathChanStrings()) {
            if (xAxis != null && xAxis.equals(s3)) {
                xExists = true;
            }
            if (yAxis != null && yAxis.equals(s3)) {
                yExists = true;
            }
        }
        if (!xExists) {
            xAxis = null;
        }
        if (!yExists) {
            yAxis = null;
        }
    }

    /**
     * 
     * @param b
     * @param undo
     */
    @Override
    public void setCursorLinked(boolean b, boolean undo) {
        super.setCursorLinked(b, undo);
        propPanel.linkCursorChanged();
    }

    /**
     * 
     * @param channel
     * @param axis
     * @param undo
     */
    public void setChannelSelection(String channel, String axis, boolean undo) {
        if (!undo) {
            if (axis.equals("x")) {
                xAxis = channel;
            } else if (axis.equals("y")) {
                yAxis = channel;
            }
            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.channelsChanged();
        } else {
            final String targetAxis = axis;
            final String oldValue;
            if (axis.equals("x")) {
                oldValue = this.xAxis;
            } else {
                oldValue = this.yAxis;
            }
            final String newValue = channel;

            final XYChartPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setChannelSelection(newValue, targetAxis, false);
                }

                @Override
                public void reverse() {
                    source.setChannelSelection(oldValue, targetAxis, false);
                }

                @Override
                public String toString() {
                    return "Channel Selection";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @param lap
     * @param value
     * @param undo
     */
    public void setLapSelection(int lap, boolean value, boolean undo) {
        if (!undo) {
            if (!value && selectedLaps.contains(new Integer(lap))) {
                selectedLaps.remove(new Integer(lap));
            } else if (value && !selectedLaps.contains(lap)) {
                selectedLaps.add(lap);
            }

            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.lapsChanged();
        } else {
            final int targetLap = lap;
            final boolean oldValue = selectedLaps.contains(lap);
            final boolean newValue = value;
            final XYChartPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setLapSelection(targetLap, newValue, false);
                }

                @Override
                public void reverse() {
                    source.setLapSelection(targetLap, oldValue, false);
                }

                @Override
                public String toString() {
                    return "Lap Selection";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public ArrayList<Integer> getSelectedLaps() {
        return this.selectedLaps;
    }

    /**
     * 
     * @param segment
     * @param value
     * @param undo
     */
    public void setSegmentSelection(int segment, boolean value, boolean undo) {
        if (!undo) {
            if (!value && selectedSegments.contains(new Integer(segment))) {
                selectedSegments.remove(new Integer(segment));
            } else if (value && !selectedSegments.contains(segment)) {
                selectedSegments.add(segment);
            }
            //Update GUI
            pluginPanel.rebuildSeries();
            propPanel.segmentsChanged();
        } else {
            final int targetSegment = segment;
            final boolean oldValue = selectedSegments.contains(segment);
            final boolean newValue = value;
            final XYChartPlugin source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.setSegmentSelection(targetSegment, newValue, false);
                }

                @Override
                public void reverse() {
                    source.setSegmentSelection(targetSegment, oldValue, false);
                }

                @Override
                public String toString() {
                    return "Segment Selection";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * 
     * @return
     */
    public ArrayList<Integer> getSelectedSegments() {
        return this.selectedSegments;
    }

    /**
     * 
     * @return
     */
    public String getXAxis() {
        return this.xAxis;
    }

    /**
     * 
     * @return
     */
    public String getYAxis() {
        return this.yAxis;
    }

    /**
     * 
     * @param title
     * @param undo
     */
    @Override
    public void setTitle(String title, boolean undo) {
        super.setTitle(title, undo);

        //Update GUI
        if (propPanel != null) {
            propPanel.titleChanged();
        }

    }

    // Methods used externally
    /**
     * 
     */
    @Override
    public void channelsChanged() {
        removeUnsupportedChannelValues();
        pluginPanel.rebuildSeries();
        propPanel.channelsChanged();

    }

    /**
     * 
     */
    @Override
    public void colorsChanged() {
        pluginPanel.rebuildSeries();

        propPanel.lapColorsChanged();
    }

    /**
     * 
     */
    @Override
    public void segmentationChanged() {
        //removeUnsupportedSegmentValues();
        pluginPanel.rebuildSeries();
        propPanel.lapsChanged();
        propPanel.segmentsChanged();
    }

    /**
     * 
     * @param position
     */
    @Override
    public synchronized void setCursorPosition(int position) {
        super.setCursorPosition(position);
        //Updating GUI
        pluginPanel.setCursorPosition(position);
    }

    /**
     * Get the main GUI component of the Plugin.
     * @return GUI component.
     */
    @Override
    public JPanel getPluginPanel() {
        return this.pluginPanel;

    }

    /**
     * Get the property GUI component of the Plugin. 
     * @return GU /**I component.
     */
    @Override
    public Component getPropertiesPanel() {
        return this.propPanel;
    }

    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        removeUnsupportedChannelValues();
        removeUnsupportedLapValues();
        removeUnsupportedSegmentValues();
        super.writeExternal(out);
        out.writeInt(this.dataVersion);
        out.writeObject(this.xAxis);
        out.writeObject(this.yAxis);
        out.writeObject(this.getSelectedLaps());
        out.writeObject(this.getSelectedSegments());
        out.writeBoolean(followDomainCursor);

    }

    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        super.readExternal(in);
        int version = in.readInt();
        if (version == 1) {
            this.xAxis = (String) in.readObject();
            this.yAxis = (String) in.readObject();
            this.selectedLaps = (ArrayList<Integer>) in.readObject();
            this.selectedSegments = (ArrayList<Integer>) in.readObject();
            this.followDomainCursor = in.readBoolean();

        }
    }
}
