package project;

import serializer.Serializer;
import channel.Channel;
import gui.MainGUI;
import java.awt.Color;
import java.io.Externalizable;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.ArrayList;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import math.Calibrator;
import math.Polynomial;
import parser.Parser;
import plugin.Plugin;
import plugin.PluginFactory;
import segmentation.Segmentation;
import view.View;
import view.ViewFactory;
import error.Error;
import error.ErrorLogger;
import java.util.HashMap;
import plugin.PluginInitException;
import view.ViewInitException;

/**
 * Class representing one project in the software.
 * @author Lauritz Møllersen
 */
public class Project implements Externalizable {
    private static final long serialVersionUID = 1L;
    private MainGUI gui;
    private ArrayList<View> views = new ArrayList<View>();
    private ArrayList<Channel> channels = new ArrayList<Channel>();
    private ArrayList<Channel> mathChannels = new ArrayList<Channel>();
    private Segmentation segmentation = null;
    private String notes = "";
    private ArrayList<Command> undoQueue;
    private ArrayList<Command> redoQueue;
    private HashMap channelColors;
    private HashMap lapColors;
    private File file;
    private int pluginCounter=1;
    private int viewCounter=1;
    private boolean saved = true;
    private int hz;
    private boolean gps = false;
    
    /**
     * Default constructor
     */
    public Project() {      
    }
    
    /**
     * Initializes a new Project, setting a gui reference and creating blank 
     * fields. Checks needs to be done in this method to see if fields
     * are already initialized (i.e. the project is being loaded from file).
     * @param gui
     */
    public void init(MainGUI gui){
    this.gui=gui;
    this.undoQueue = new ArrayList<Command>();
    this.redoQueue = new ArrayList<Command>();
    this.hz = 100;
    if(channelColors==null) channelColors = new HashMap();
    if(lapColors==null) lapColors = new HashMap();
    if(views.size()==0){
        createNewView(view.basic.BasicView.class, false);
        
    } else {
    for(View v: views){
            try {
                v.init(this);
                gui.addView(v, views.indexOf(v));
            } catch (ViewInitException ex) {
                   ErrorLogger.addError(new Error("View error", "View failed to initialize. "
                    + "View initialization aborted."));
            }
        
    }}
}
    /**
     * Returns the project's state (saved or not saved).
     * @return isSaved
     */
    public boolean isSaved(){
        return saved;
    }
    /**
     * Gets a the channel color for a given channel.
     * @param channel Channel name
     * @return Color
     */
    public Color getChannelColor(String channel){
       if(channelColors.containsKey(channel))return (Color)channelColors.get(channel);
       channelColors.put(channel, generateChannelColor(channel));
       return (Color)channelColors.get(channel);
   }
   
   /**
    * Sets the channel color for a given channel.
    * @param channel Target channel
    * @param color Channel color
    * @param undo
    */
   public void setChannelColor(String channel, Color color, boolean undo) { // UNDO
       if(!undo){
           channelColors.remove(channel);
       
       channelColors.put(channel, color);
       //Notify plugins
       for(View v: views){
           for(Plugin p: v.getPlugins()){
               p.colorsChanged();
           }
       }}
       else { 
           final Color oldValue = this.getChannelColor(channel);
           final Color newValue = color;
           final String targetChannel = channel;
           final Project source = this; 
           Command c = new Command() {
                @Override
                public void execute() {
                    source.setChannelColor(targetChannel, newValue, false);
                }

                @Override
                public void reverse() {
                    source.setChannelColor(targetChannel, oldValue, false);
                }

                @Override
                public String toString() {
                    return "Change Channel Color";
                }
            };
            c.execute();
            addUndoHistory(c);
           
       }
   }
   
   private Color generateChannelColor(String channel){
       int index = channels.indexOf(getChannel(channel));
       if(index==-1) index = mathChannels.indexOf(getChannel(channel));
       if(index%6==0)return Color.BLUE;
       if(index%5==0)return Color.RED;
       if(index%4==0)return Color.DARK_GRAY;
       if(index%3==0)return Color.MAGENTA;
       if(index%2==0)return Color.ORANGE;
       return Color.LIGHT_GRAY;
   }
   /**
    * Returns the global channel colors.
    * @return HashMap containing color values
    */
   public HashMap getChannelColors(){
       return this.channelColors;
   }
   
   /**
    * Returns the color of a given lap
    * @param lap Lap index
    * @return Color
    */
   public Color getLapColor(int lap){
       if(lapColors.containsKey(lap))return (Color)lapColors.get(lap);
       lapColors.put(lap,generateLapColor(lap));
       return (Color)lapColors.get(lap);
   }
   /**
    * Returns the global lap colors.
    * @return HashMap containing color values
    */
   public HashMap getLapColors(){
       return this.lapColors;
   }
   
   /**
    * Sets the lap color for a given lap
    * @param lap Target lap
    * @param color Lap color
    * @param undo
    */
   public void setLapColor(int lap, Color color, boolean undo){ 
       if(!undo){
       lapColors.remove(lap);
       lapColors.put(lap, color);
       
       //Notify plugins
       for(View v: views){
           for(Plugin p: v.getPlugins()){
               p.colorsChanged();
           }
       }}else {
           
           final Color oldValue = this.getLapColor(lap);
           final Color newValue = color;
           final int targetLap = lap;
           final Project source = this; 
           Command c = new Command() {
                @Override
                public void execute() {
                    source.setLapColor(targetLap, newValue, false);
                }

                @Override
                public void reverse() {
                    source.setLapColor(targetLap, oldValue, false);
                }

                @Override
                public String toString() {
                    return "Change Lap Color";
                }
            };
            c.execute();
            addUndoHistory(c);
       }
   }
   
   private Color generateLapColor(int lap){
       if(lap%6==0)return Color.BLUE;
       if(lap%5==0)return Color.RED;
       if(lap%4==0)return Color.DARK_GRAY;
       if(lap%3==0)return Color.MAGENTA;
       if(lap%2==0)return Color.ORANGE;
       return Color.LIGHT_GRAY;
   }
    /**
     * Method to undo the last registered property change
     */
   public void undo() {
        undoQueue.get(undoQueue.size() - 1).reverse();
        Command c = undoQueue.remove(undoQueue.size() - 1);
        redoQueue.add(c);
        gui.setRedoEnabled(true);
        gui.setRedoText("Redo "+c.toString());
        if (undoQueue.isEmpty()) {
            gui.setUndoEnabled(false);
            gui.setUndoText("Undo");
        } else gui.setUndoText("Undo "+undoQueue.get(undoQueue.size() - 1).toString());
    }

   /**
    * Method to redo a property change that has been undone.
    */
   public void redo() {
        redoQueue.get(redoQueue.size() - 1).execute();
        Command c = redoQueue.remove(redoQueue.size() - 1);
        undoQueue.add(c);
        gui.setUndoEnabled(true);
        gui.setUndoText("Undo "+c.toString());
        if (redoQueue.isEmpty()) {
            gui.setRedoEnabled(false);
            gui.setRedoText("Redo");
        } else gui.setRedoText("Redo "+redoQueue.get(redoQueue.size() - 1).toString());
    }
   /**
    * Sets the saved state of the project
    * @param saved State
    */
   public void setSaved(boolean saved){
        this.saved = saved;
    }
    /**
     * Register a property change in the project undo history.
    * @param c 
    */
    public void addUndoHistory(Command c) {
        saved = false;
        undoQueue.add(c);
        redoQueue = new ArrayList<Command>();
        gui.setUndoEnabled(true);
        gui.setRedoEnabled(false);
        gui.setUndoText("Undo "+c.toString());
        gui.setRedoText("Redo");
    }
  
    /**
     * Imports data from a csv file to a set of Channel objects.
     * @param path The path of the csv file.
     * @param prop  
     */
    public void importData(String path, Properties prop) {
        setChannels(Parser.parseCSV(path, prop));
    }
  
    /**
     * Method to create a new view in the current project.
     * @param type Type of the new view.
     * @param undo  
     */
    public void createNewView(Class type, boolean undo) {
        View v = ViewFactory.createNewView(type, viewCounter);
        if(v!=null){
        
            try {
                v.init(this);
                 this.addView(v,getViews().size(),undo);
                 viewCounter++;
            } catch (ViewInitException ex) {
             ErrorLogger.addError(new Error("View error", "View failed to initialize. "
                    + "View initialization aborted."));
            }
        }
        else {
            //Something went wrong in instantiation.
        }
    }


    /**
     * Method to create a new plugin in the current view.
     * @param type Type of the new view.
     * @param undo 
     * @param targetView Target view for the plugin to be added in.
     */
    public void createNewPlugin(Class type, View targetView, boolean undo) {
        Plugin p = PluginFactory.createPlugin(type, pluginCounter);
        if(p!=null){
        try {
            p.init(targetView);
            targetView.registerPlugin(p, undo);
            pluginCounter++;
        } catch (PluginInitException ex) {
            //Something went wrong in init
           // Logger.getLogger(Project.class.getName()).log(Level.SEVERE, null, ex);
        }
       
        } else {
            //Something went wrong in instantiation
        }
    }
   
    
    /**
     * Loads a view from file and adds it to the project
     * @param file Input File
     */
    public void loadViewFromFile(File file){
        
        try{    
           View v = ViewFactory.deserializeFromFile(file); 
           v.init(this);
           this.addView(v, getViews().size(), false);  
        }
        catch (ViewInitException ex) {
              ErrorLogger.addError(new Error("View error", "View failed to initialize. "
                    + "View initialization aborted."));
        }           catch (IOException ex) {
            ErrorLogger.addError(new Error("File error", "View file format not recognized."));
        } catch (ClassNotFoundException ex) {
          
        }
    }
    
    /**
     * Method to add a {@link View} (tab) to the project.
     * @param v The view object to be added.
     * @param viewPosition 
     * @param undo  
     */
    public void addView(View v, int viewPosition, boolean undo) {
        if(!undo){
            views.add(viewPosition,v);
            getGui().addView(v, viewPosition);
        } 
            else { 
           final View targetView = v;
           final int position = viewPosition;
           
           final Project source = this; 
           Command c = new Command() {
                @Override
                public void execute() {
                    source.addView(targetView, position, false);
                }

                @Override
                public void reverse() {
                    source.removeView(targetView, false);
                }

                @Override
                public String toString() {
                    return "Add View";
                }
            };
            c.execute();
            addUndoHistory(c);
           
       
        }
       
    }

    /**
     * @return the views
     */
    public ArrayList<View> getViews() {
        return views;
    }

    /**
     * @param views the views to set
     */
    public void setViews(ArrayList<View> views) {
        this.views = views;
    }
      
    /**
     * Returns a channel with the given name
     * @param name Channel name
     * @return Channel
     */
    public Channel getChannel(String name) {
        Channel tempChan;
        for(int i = 0; i < getChannels().size(); i++) {
            tempChan = getChannels().get(i);
            if(tempChan.getName().equals(name)) return tempChan;
        }
        for(int i = 0; i < getMathChannels().size(); i++) {
            tempChan = getMathChannels().get(i);
            if(tempChan.getName().equals(name)) return tempChan;
        }
        ErrorLogger.addError(new Error(0, "Input error", "Channel of name " + name + " not found."));
        return null;
    }
    
    /**
     * Returns a channel of the given type
     * @param type Channel type
     * @return Channel
     */
    public Channel getChannelType(int type) {
        Channel tempChan;
        for(int i = 0; i < getChannels().size(); i++) {
            tempChan = getChannels().get(i);
            if(tempChan.getType() == type) return tempChan;
        }
        ErrorLogger.addError(new Error(0, "Input error", "Channel of type #" + type + " not found."));
        return null;
    }
    
     /**
      * Sets the channel type of a given channel
      * @param number Channel index
      * @param type Channel type
      */
     public void setChannelType(int number, int type) {
        if(number >= 0 && number < getChannels().size()) {
            getChannels().get(number).setType(type);
        }
        else {
            ErrorLogger.addError(new Error(0, "Input error", "Channel not found/Invalid channel ID."));
        }
    }
    
     /**
      * Returns the amount of channels available
      * @return Amount of channels
      */
     public int getNumChannels() {
        return getChannels().size();
    }
    
    /**
     * 
     * @param polys
     * @deprecated
     */
    @Deprecated
    public void calibrateChannels(Polynomial polys) {
        Channel tempChannel;
        double[] tempValues;
        
        for(int i = 0; i < getChannels().size(); i++) {
            tempChannel = getChannels().get(i);
            if(tempChannel.getType() <= 0) {
                tempValues = tempChannel.getValues();
                Calibrator.calibrate(tempChannel);
            }
        }
    }

    /**
     * Returns a reference to the main GUI
     * @return GUI reference
     */
    public MainGUI getGui(){
        return gui;
        
    }
   

    /**
     * @param gui the gui to set
     */
    public void setGui(MainGUI gui) {
        this.gui = gui;
    }

    /**
     * @return the segmentation
     */
    public Segmentation getSegmentation() {
        return segmentation;
    }

    /**
     * @param segmentation the segmentation to set
     */
    public void setSegmentation(Segmentation segmentation) {
        this.segmentation = segmentation;
        for(View v: views){
            for(Plugin p: v.getPlugins()){
                p.segmentationChanged();
            }
        }
    }

    /**
     * @return the channels
     */
    public ArrayList<Channel> getChannels() {
        return channels;
    }

    /**
     * @param channels the channels to set
     */
    public void setChannels(ArrayList<Channel> channels) {
        this.channels = channels;
        for(View v: views){

                //notify plugins
       for(Plugin p: v.getPlugins()){
           p.channelsChanged();
       }
    }

    }
    
    /**
     * Returns an array containing all channel names
     * @return String array
     */
    public String[] getChanStrings() {
        final int size = channels.size();
        String[] temp = new String[size];
        Channel tempChan;
        for (int i = 0; i < size; i++) {
            tempChan = channels.get(i);
            temp[i] = tempChan.getName();
        }
        return temp;
    }

    /**
     * Returns an array containing all math channel names
     * @return String array
     */
    public String[] getMathChanStrings() {
        final int size = mathChannels.size();
        String[] temp = new String[size];
        Channel tempChan;
        for (int i = 0; i < size; i++) {
            tempChan = mathChannels.get(i);
            temp[i] = tempChan.getName();
        }
        return temp;
    }
    
    /**
     * @return the hz
     */
    public int getHz() {
        return hz;
    }
    
    /**
     * Set the hz
     * @param hz
     */
    public void setHz(int hz) {
        this.hz = hz;
    }
    
    /**
     * Return a list of mathchannels
     * @return 
     */
    public ArrayList<Channel> getMathChannels() {
        return mathChannels;
    }
    
    /**
     * Sets the set of math channels
     * @param chans
     */
    public void setMathChannels(Channel[] chans) {
        ArrayList<Channel> newChans = new ArrayList<Channel>();
        Channel temp = null;
        for(int i = 0; i < chans.length; i++) {
            temp = chans[i];
            if(temp != null) newChans.add(temp);
        }
        mathChannels = newChans;
        
        //notify plugins
        for(View v: views){
  
       for(Plugin p: v.getPlugins()){
           p.channelsChanged();
       }
    }
    }
    
    /**
     * Returns an array containing the data channels
     * @return
     */
    public Channel[] getChannelsArray() {
        final int size = channels.size();
        Channel[] tempArray = new Channel[size];
        for(int i = 0; i < size; i++) {
            tempArray[i] = channels.get(i);
        }
        return tempArray;
    }

    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        ArrayList<byte[]> viewList = new ArrayList<byte[]>();
        for (View v : this.views){
            viewList.add(ViewFactory.serializeToByteArray(v));
        }
        out.writeObject(viewList);
        out.writeInt(this.hz);
        out.writeObject(this.channels);
        out.writeObject(this.segmentation);
        out.writeInt(this.pluginCounter);
        out.writeInt(this.viewCounter);
        out.writeObject(this.notes);
        
    }

    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        ArrayList<byte[]> viewList = (ArrayList<byte[]>)in.readObject();
        for (byte[] b : viewList){
            this.views.add(ViewFactory.deserializeFromByteArray(b));  
        }
        this.hz = in.readInt();
        this.channels = (ArrayList<Channel>)in.readObject();
        this.segmentation = (Segmentation)in.readObject(); 
        this.pluginCounter = in.readInt();
        this.viewCounter = in.readInt();
        this.notes = (String) in.readObject();
    }
    
    /**
     * Loads a project from file
     * @param file Input File
     * @return Project
     */
    public static Project loadFromFile(File file){
        Project p;
        try {
            p = (Project)Serializer.deserializeFromFile(file);
             p.setFile(file);
            return p;
        } catch (IOException ex) {
               ErrorLogger.addError(new Error("File error", "Project file format not recognized."));
        } catch (ClassNotFoundException ex) {
            ex.printStackTrace();
            ErrorLogger.addError(new Error("Fatal error", "Project class not found."));
        }
        return null;
    }
    
    /**
     * Savees a project to a file
     * @param file Target File
     */
    public void saveToFile(File file){
        saved = true;
        try {
            Serializer.serializeToFile(this, file);
            this.setFile(file);
        } catch (IOException ex) {
            Logger.getLogger(Project.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    
    /**
     * 
     * @return
     */
    public boolean isGps() {
        return gps;
    }
    
    /**
     * 
     * @param gps
     */
    public void setGps(boolean gps) {
        this.gps = gps;
    }

    /**
     * Saves a view to file
     * @param currentView View to save
     * @param selectedFile Target file
     */
    public void saveViewToFile(View currentView, File selectedFile) {
        try {
            ViewFactory.serializeToFile(currentView, selectedFile);
        } catch (IOException ex) {
        }
    }

    /**
     * Removes a view from the project
     * @param v View to remove
     * @param undo
     */
    public void removeView(View v, boolean undo) {
        if(!undo){
               for(Plugin p: v.getPlugins()){
                  // Component focusedPlugin = gui.getCurrentPluginPropComponent();
                  Plugin focusedPlugin = gui.getCurrentPlugin();
                   if(focusedPlugin!=null && focusedPlugin.equals(p))gui.setCurrentPlugin(null);//}
            getGui().removeView(v);
        }}
            else { 
           final View targetView = v;
           final int viewPosition = views.indexOf(v);
           
           final Project source = this; 
           Command c = new Command() {
                @Override
                public void execute() {
                    source.removeView(targetView, false);
                }

                @Override
                public void reverse() {
                    source.addView(targetView, viewPosition, false);
                }
                @Override
                public String toString() {
                    return "Delete View";
                }
            };
            c.execute();
            addUndoHistory(c);
        }
    }

    /**
     * Returns the undo history of the project
     * @return Undo History
     */
    public ArrayList<Command> getUndoHistory() {
        return this.undoQueue;
    }
    
    /**
     * @return the notes
     */
    public String getNotes() {
        return notes;
    }

    /**
     * @param notes the notes to set
     */
    public void setNotes(String notes) {
        this.notes = notes;
    }

    /**
     * Sets the file associated with the project
     * @param file Target File
     */
    public void setFile(File file) {
        this.file = file;
        if(gui!=null){
                gui.setFrameTitle("Revolve Analyzer - "+file.toString());
    }}
    
    /**
     * Gets the file associated with the project
     * @return File
     */
    public File getFile(){
        return this.file;
    }

    

}
