package view;

import java.awt.Component;
import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.ArrayList;
import plugin.PluginInitException;
import project.Project;
import plugin.Plugin;
import plugin.PluginFactory;
import project.Command;

/**
 *
 * @author Per Øyvind
 */
public class DefaultView implements Externalizable, view.View {

    private static final long serialVersionUID = 1L;
    private Project project;
    private String title;
    private ArrayList<Plugin> plugins = new ArrayList<Plugin>();
    private int cursorPosition;

    /**
     * Initializes the view by performing necessary operations
     * and setting the parent project reference. This method needs 
     * to check if the view's fields are already set (i.e. the view
     * is loaded from disk).
     * @param p The parent project of the view.
     * @throws ViewInitException Exception used to abort the initialization
     * of the view
     */
    @Override
    public void init(Project p) throws ViewInitException {
        this.project = p;
        this.cursorPosition = 0;
        for (int i = 0; i < plugins.size(); i++) {
            Plugin plugin = plugins.get(i);
            if (plugin != null) //plugin class not found
            {
                try {
                    plugin.init(this);
                } catch (PluginInitException ex) {
                    throw new ViewInitException();
                }
            } else {
                plugins.remove(i);
                i--;
            }
        }
    }

    /**
     * Sets the title of the view.
     * This method is only used by the ViewFactory for initialization purposes.
     * @param title
     */
    public void setTitle(String title) {
        this.title = title;
    }

    /**
     * Registers a plugin to the view.
     * @param p Plugin to register.
     * @param undo  
     */
    @Override
    public void registerPlugin(Plugin p, boolean undo) {
        if (!undo) {
            plugins.add(p);
        } else {
            final Plugin plugin = p;

            final DefaultView source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.registerPlugin(plugin, false);
                }

                @Override
                public void reverse() {
                    source.removePlugin(plugin, false);
                }

                @Override
                public String toString() {
                    return "Add Plugin";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }

    }

    /**
     * Returns the list of plugins contained in the view
     * @return List of plugins
     */
    @Override
    public void removePlugin(Plugin p, boolean undo) {
        if (!undo) {
            plugins.remove(p);
            p.stopPlayback();
            Plugin currentPlugin = getParentProject().getGui().getCurrentPlugin();
            if (currentPlugin != null && currentPlugin.equals(p)) {
                getParentProject().getGui().setCurrentPlugin(null);
            }
        } else {
            final Plugin plugin = p;

            final DefaultView source = this;
            Command c = new Command() {

                @Override
                public void execute() {
                    source.removePlugin(plugin, false);
                }

                @Override
                public void reverse() {
                    source.registerPlugin(plugin, false);
                }

                @Override
                public String toString() {
                    return "Remove Plugin";
                }
            };
            c.execute();
            getParentProject().addUndoHistory(c);
        }
    }

    /**
     * Returns the current title of the view
     * @return The view title
     */
    @Override
    public String getTitle() {
        return title;
    }

    /**
     * Sets the title of the view
     * @param title The view title
     * @param undo
     */
    @Override
    public void setTitle(String title, boolean undo) {
        if (!undo) {
            this.title = title;
            getParentProject().getGui().viewTitlesChanged();
        } else {
            if (!title.equals(this.getTitle())) {
                final String oldValue = this.getTitle();
                final String newValue = title;
                final DefaultView source = this;
                Command c = new Command() {

                    @Override
                    public void execute() {
                        source.setTitle(newValue, false);
                    }

                    @Override
                    public void reverse() {
                        source.setTitle(oldValue, false);
                    }

                    @Override
                    public String toString() {
                        return "Set View Title";
                    }
                };
                c.execute();
                getParentProject().addUndoHistory(c);
            }

        }

    }

    /**
     * Get the parent project of the view.
     * @return The parent project.
     */
    @Override
    public Project getParentProject() {
        return this.project;
    }

    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeObject(this.title);
        out.writeObject(serializePlugins());
    }

    @Override
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        this.title = (String) in.readObject();
        ArrayList<byte[]> list = (ArrayList<byte[]>) in.readObject();

        for (byte[] b : list) {
            this.plugins.add(PluginFactory.deserializeFromByteArray(b));
        }

    }

    /**
     * Returns the list of plugins contained in the view
     * @return List of plugins
     */
    @Override
    public ArrayList<Plugin> getPlugins() {
        return this.plugins;
    }

    /**
     * Get the main GUI component of the View.
     * @return GUI component.
     */
    @Override
    public Component getViewPanel() {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    /**
     * Get the property GUI component of the View.
     * @return GUI component.
     */
    @Override
    public Component getPropertiesPanel() {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    private ArrayList<byte[]> serializePlugins() {
        ArrayList<byte[]> list = new ArrayList<byte[]>();
        for (Plugin p : plugins) {
            list.add(PluginFactory.serializeToByteArray(p));
        }
        return list;
    }

    /**
     * Method use to set the view's cursor position and update
     * any linked plugins.
     * @param lineNumber The position to set the cursor to
     */
    public synchronized void setGlobalCursorPosition(int lineNumber) {
        this.cursorPosition = lineNumber;
        for (Plugin p : getPlugins()) {
            if (p.isCursorLinked()) {
                p.setCursorPosition(lineNumber);
            }
        }
    }

    /**
     * Returns the view's cursor position.
     * @return Cursor position
     */
    @Override
    public int getGlobalCursorPosition() {
        return cursorPosition;
    }

    /**
     * Method used to update the plugin titles in the view
     */
    @Override
    public void pluginTitlesChanged() {
    }
}
